@@ -3,18 +3,13 @@ from collections import namedtuple
 from threading import RLock
 
 from mo_future import text
+from mo_imports import export, expect
 
-from mo_parsing.cache import packrat_cache
-from mo_parsing.engine import Engine
-from mo_parsing.exceptions import (
-    ParseException,
-    ParseFatalException,
-    conditionAsParseAction,
-)
+from mo_parsing import whitespaces
+from mo_parsing.exceptions import ParseException
 from mo_parsing.results import ParseResults
-from mo_parsing.utils import Log, MAX_INT, wrap_parse_action, empty_tuple, is_forward
+from mo_parsing.utils import Log, MAX_INT, wrap_parse_action, empty_tuple
 
-# import later
 (
     SkipTo,
     Many,
@@ -23,19 +18,35 @@ from mo_parsing.utils import Log, MAX_INT, wrap_parse_action, empty_tuple, is_fo
     Optional,
     NotAny,
     Suppress,
-    _flatten,
-    replaceWith,
-    quotedString,
     And,
     MatchFirst,
     Or,
-    Each,
+    MatchAll,
     Empty,
     StringEnd,
     Literal,
     Token,
     Group,
-) = [None] * 19
+    regex_parameters,
+) = expect(
+    "SkipTo",
+    "Many",
+    "ZeroOrMore",
+    "OneOrMore",
+    "Optional",
+    "NotAny",
+    "Suppress",
+    "And",
+    "MatchFirst",
+    "Or",
+    "MatchAll",
+    "Empty",
+    "StringEnd",
+    "Literal",
+    "Token",
+    "Group",
+    "regex_parameters",
+)
 
 DEBUG = False
 
@@ -53,6 +64,7 @@ def add_reset_action(action):
 
 
 locker = RLock()
+streamlined = {}
 
 
 def entrypoint(func):
@@ -64,212 +76,36 @@ def entrypoint(func):
                 except Exception as e:
                     Log.error("reset action failed", cause=e)
 
-            self = args[0]
-            if not self.streamlined and (
-                not is_forward(self) or not self.expr.streamlined
-            ):
-                Log.alert("Expecting expression to be streamlined before use")
-                self = self.streamline()
-            return func(self, *args[1:], **kwargs)
+            return func(*args, **kwargs)
 
     return output
 
 
-class ParserElement(object):
-    """Abstract base level parser element class."""
-
-    __slots__ = [
-        "parseAction",
-        "parser_name",
-        "token_name",
-        "engine",
-        "streamlined",
-        "min_cache",
-        "parser_config",
-    ]
-    Config = namedtuple("Config", ["callDuringTry", "failAction", "lock_engine"])
-
-    def __init__(self):
-        self.parseAction = list()
-        self.parser_name = ""
-        self.token_name = None
-        self.engine = engine.CURRENT
-        self.streamlined = False
-        self.min_cache = -1
-
-        self.parser_config = self.Config(*([None] * len(self.Config._fields)))
-        self.set_config(callDuringTry=False, failAction=None, lock_engine=None)
-
-    def set_config(self, **map):
-        data = {
-            **dict(zip(self.parser_config.__class__._fields, self.parser_config)),
-            **map,
-        }
-        self.parser_config = self.Config(*(data[f] for f in self.Config._fields))
-
-    def copy(self):
-        output = object.__new__(self.__class__)
-        le = self.parser_config.lock_engine
-        output.engine = le or engine.CURRENT
-        output.parseAction = self.parseAction[:]
-        output.parser_name = self.parser_name
-        output.token_name = self.token_name
-        output.parser_config = self.parser_config
-        output.streamlined = self.streamlined
-        output.min_cache = -1
-        return output
-
-    def set_parser_name(self, name):
-        """
-        Define name for this expression, makes debugging and exception messages clearer.
-
-        Example::
-
-            Word(nums).parseString("ABC")  # -> Exception: Expected W:(0123...) (at char 0), (line:1, col:1)
-            Word(nums).set_parser_name("integer").parseString("ABC")  # -> Exception: Expected integer (at char 0), (line:1, col:1)
-        """
-        self.parser_name = name
-        return self
-
-    def setBreak(self, breakFlag=True):
-        """Method to invoke the Python pdb debugger when this element is
-        about to be parsed. Set ``breakFlag`` to True to enable, False to
-        disable.
-        """
-        if breakFlag:
-            _parseMethod = self._parse
-
-            def breaker(string, loc, doActions=True):
-                import pdb
-
-                # this call to pdb.set_trace() is intentional, not a checkin error
-                pdb.set_trace()
-                return _parseMethod(string, loc, doActions)
-
-            breaker._originalParseMethod = _parseMethod
-            self._parse = breaker
-        else:
-            if hasattr(self._parse, "_originalParseMethod"):
-                self._parse = self._parse._originalParseMethod
-        return self
-
-    def clearParseAction(self):
-        """
-        Add one or more parse actions to expression's list of parse actions. See `setParseAction`.
-
-        See examples in `copy`.
-        """
-        output = self.copy()
-        output.parseAction = []
-        return output
-
-    def addParseAction(self, *fns, callDuringTry=False):
-        """
-        Add one or more parse actions to expression's list of parse actions. See `setParseAction`.
-
-        See examples in `copy`.
-        """
-        output = self.copy()
-        output.parseAction += list(map(wrap_parse_action, fns))
-        output.set_config(
-            callDuringTry=self.parser_config.callDuringTry or callDuringTry
-        )
-        return output
-
-    def addCondition(
-        self, *fns, message=None, fatal=False, callDuringTry=False, **kwargs
-    ):
-        """Add a boolean predicate function to expression's list of parse actions. See
-        `setParseAction` for function call signatures. Unlike ``setParseAction``,
-        functions passed to ``addCondition`` need to return boolean success/fail of the condition.
-
-        Optional keyword arguments:
-        - message = define a custom message to be used in the raised exception
-        - fatal   = if True, will raise ParseFatalException to stop parsing immediately; otherwise will raise ParseException
-
-        """
-        output = self.copy()
-        for fn in fns:
-            output.parseAction.append(conditionAsParseAction(
-                fn, message=message, fatal=fatal
-            ))
-
-        output.set_config(
-            callDuringTry=self.parser_config.callDuringTry or callDuringTry
-        )
-        return output
-
-    def setFailAction(self, fn):
-        """Define action to perform if parsing fails at this expression.
-        Fail acton fn is a callable function that takes the arguments
-        ``fn(s, loc, expr, err)`` where:
-        - expr = the parse expression that failed
-        - loc = location where expression match was attempted and failed
-        - s = string being parsed
-        - err = the exception thrown
-        The function returns no value.  It may throw `ParseFatalException`
-        if it is desired to stop parsing immediately."""
-        self.set_config(failAction=fn)
-        return self
-
-    def is_annotated(self):
-        return self.parseAction or self.token_name or self.parser_name
-
-    def min_length(self):
-        if not hasattr(self, "min_cache"):
-            Log.error("should not happen")
-        if self.min_cache >= 0:
-            return self.min_cache
-        min_ = self._min_length()
-        if self.streamlined:
-            self.min_cache = min_
-        return min_
-
-    def _min_length(self):
-        return 0
-
-    def parseImpl(self, string, start, doActions=True):
-        return ParseResults(self, start, start, [])
-
-    def _parse(self, string, start, doActions=True):
-        lookup = (self, string, start, doActions)
-        value = packrat_cache.get(lookup)
-        if value is not None:
-            if isinstance(value, Exception):
-                raise value
-            return value
-
+class Parser(object):
+    def __init__(self, element):
+        self.element = element = element.streamline()
         try:
-            index = self.engine.skip(string, start)
-            try:
-                tokens = self.parseImpl(string, index, doActions)
-            except Exception as cause:
-                self.parser_config.failAction and self.parser_config.failAction(
-                    self, start, string, cause
-                )
-                raise
+            engs = element.whitespace
+            while isinstance(engs, list):
+                engs = [e for e in engs if e is not None]
+                if not engs:
+                    break
+                whitespace = engs[0]
+                if any(e.id != whitespace.id for e in engs[1:]):
+                    Log.error("must dis-ambiguate the whitespace before parsing")
+                engs = whitespace
 
-            if self.parseAction and (doActions or self.parser_config.callDuringTry):
-                for fn in self.parseAction:
-                    tokens = fn(tokens, index, string)
-        except ParseException as cause:
-            packrat_cache.set(lookup, cause)
-            raise
+            self.whitespace = (
+                engs or whitespaces.CURRENT or whitespaces.STANDARD_WHITESPACE
+            )
+        except Exception as cause:
+            Log.error("problem", cause=cause)
 
-        packrat_cache.set(lookup, tokens)
-        return tokens
+        with self.whitespace:
+            self.element = Group(element)
 
-    def tryParse(self, string, start):
-        try:
-            return self._parse(string, start, doActions=False).end
-        except ParseFatalException as cause:
-            raise ParseException(self, start, string, cause=cause)
-
-    def canParseNext(self, string, start):
-        try:
-            return self.tryParse(string, start)
-        except ParseException:
-            return False
+        self.named = bool(element.token_name)
+        self.streamlined = True
 
     @entrypoint
     def parseString(self, string, parseAll=False):
@@ -296,59 +132,59 @@ class ParserElement(object):
         - explicitly expand the tabs in your input string before calling ``parseString``.
 
         """
-        cache.resetCache()
-        expr = self.streamline()
-        for e in expr.engine.ignore_list:
-            e.streamline()
-        if expr.token_name:
-            # TOP LEVEL NAMES ARE NOT ALLOWED
-            expr = Group(expr)
-        tokens = expr._parse(string, 0)
-        end = tokens.end
+        return self._parseString(string, parseAll=parseAll)
+
+    def _parseString(self, string, parseAll=False):
+        start = self.whitespace.skip(string, 0)
+        tokens = self.element._parse(string, start)
         if parseAll:
-            end = expr.engine.skip(string, end)
+            end = self.whitespace.skip(string, tokens.end)
             StringEnd()._parse(string, end)
-        return tokens
+
+        if self.named:
+            return tokens
+        else:
+            return tokens.tokens[0]
 
     @entrypoint
     def scanString(self, string, maxMatches=MAX_INT, overlap=False):
         """
-        Scan the input string for expression matches.  Each match will return the
-        matching tokens, start location, and end location.  May be called with optional
-        ``maxMatches`` argument, to clip scanning after 'n' matches are found.  If
-        ``overlap`` is specified, then overlapping matches will be reported.
-
-        Note that the start and end locations are reported relative to the string
-        being parsed.  See `parseString` for more information on parsing
-        strings with embedded tabs.
+        :param string: TO BE SCANNED
+        :param maxMatches: MAXIMUM NUMBER MATCHES TO RETURN
+        :param overlap: IF MATCHES CAN OVERLAP
+        :return: SEQUENCE OF ParseResults, start, end
         """
-        if not self.streamlined:
-            self.streamline()
-            for e in self.engine.ignore_list:
-                e.streamline()
+        return (
+            (t.tokens[0], s, e)
+            for t, s, e in self._scanString(
+                string, maxMatches=maxMatches, overlap=overlap
+            )
+        )
 
+    def _scanString(self, string, maxMatches=MAX_INT, overlap=False):
         instrlen = len(string)
-        end = 0
-        cache.resetCache()
+        start = end = 0
         matches = 0
         while end <= instrlen and matches < maxMatches:
-            start = self.engine.skip(string, end)
             try:
-                tokens = self._parse(string, start)
+                start = self.whitespace.skip(string, end)
+                tokens = self.element._parse(string, start)
             except ParseException:
                 end = start + 1
             else:
                 matches += 1
-                yield tokens, start, tokens.end
+                yield tokens, tokens.start, tokens.end
                 if overlap or tokens.end <= end:
                     end += 1
                 else:
                     end = tokens.end
 
+    @entrypoint
     def transformString(self, string):
         """
-        Extension to `scanString`, to modify matching text with modified tokens that may
-        be returned from a parse action.  To use ``transformString``, define a grammar and
+        Modify matching text with results of a parse action.
+
+        To use ``transformString``, define a grammar and
         attach a parse action to it that modifies the returned token list.
         Invoking ``transformString()`` on a target string will then scan for matches,
         and replace the matched text patterns according to the logic in the parse
@@ -365,12 +201,16 @@ class ParserElement(object):
 
             Now Is The Winter Of Our Discontent Made Glorious Summer By This Sun Of York.
         """
+        return self._transformString(string)
+
+    def _transformString(self, string):
         out = []
-        lastE = 0
+        end = 0
         # force preservation of <TAB>s, to minimize unwanted transformation of string, and to
         # keep string locs straight between transformString and scanString
-        for t, s, e in self.scanString(string):
-            out.append(string[lastE:s])
+        for t, s, e in self._scanString(string):
+            out.append(string[end:s])
+            t = t.tokens[0]
             if t:
                 if isinstance(t, ParseResults):
                     out.append("".join(t))
@@ -378,35 +218,30 @@ class ParserElement(object):
                     out.append("".join(t))
                 else:
                     out.append(t)
-            lastE = e
-        out.append(string[lastE:])
+            end = e
+        out.append(string[end:])
         out = [o for o in out if o]
-        return "".join(map(text, _flatten(out)))
+        return "".join(map(text, out))
 
+    @entrypoint
     def searchString(self, string, maxMatches=MAX_INT):
         """
-        Another extension to `scanString`, simplifying the access to the tokens found
-        to match the given parse expression.  May be called with optional
-        ``maxMatches`` argument, to clip searching after 'n' matches are found.
+        :param string: Content to scan
+        :param maxMatches: Limit number of matches
+        :return: All the matches, packaged as ParseResults
         """
+        return self._searchString(string, maxMatches=maxMatches)
 
-        if isinstance(self, Group):
-            g = self
-            scanned = [t for t, s, e in self.scanString(string, maxMatches)]
-        else:
-            g = Group(self)
-            scanned = [
-                ParseResults(g, s, e, [t])
-                for t, s, e in self.scanString(string, maxMatches)
-            ]
-
+    def _searchString(self, string, maxMatches=MAX_INT):
+        scanned = [t for t, s, e in self._scanString(string, maxMatches)]
         if not scanned:
-            return ParseResults(ZeroOrMore(g), -1, -1, [])
+            return ParseResults(ZeroOrMore(self.element), -1, 0, [])
         else:
             return ParseResults(
-                ZeroOrMore(g), scanned[0].start, scanned[-1].end, scanned
+                ZeroOrMore(self.element), scanned[0].start, scanned[-1].end, scanned
             )
 
+    @entrypoint
     def split(self, string, maxsplit=MAX_INT, includeSeparators=False):
         """
         Generator method to split a string using the given expression as a separator.
@@ -423,15 +258,246 @@ class ParserElement(object):
 
             ['This', ' this', '', ' this sentence', ' is badly punctuated', '']
         """
-        splits = 0
+        return self._split(
+            string, maxsplit=maxsplit, includeSeparators=includeSeparators
+        )
+
+    def _split(self, string, maxsplit=MAX_INT, includeSeparators=False):
         last = 0
-        for t, s, e in self.scanString(string, maxMatches=maxsplit):
+        for t, s, e in self._scanString(string, maxMatches=maxsplit):
             yield string[last:s]
             if includeSeparators:
-                yield t[0]
+                yield t.tokens[0]
             last = e
         yield string[last:]
 
+
+class ParserElement(object):
+    """Abstract base level parser element class."""
+
+    zero_length = False
+    __slots__ = [
+        "parseAction",
+        "parser_name",
+        "token_name",
+        "streamlined",
+        "min_length_cache",
+        "parser_config",
+    ]
+    Config = namedtuple("Config", ["callDuringTry", "failAction"])
+
+    def __init__(self):
+        self.parseAction = list()
+        self.parser_name = ""
+        self.token_name = ""
+        self.streamlined = False
+        self.min_length_cache = -1
+
+        self.parser_config = self.Config(*([None] * len(self.Config._fields)))
+        self.set_config(callDuringTry=False, failAction=None, lock_engine=None)
+
+    def set_config(self, **map):
+        data = {
+            **dict(zip(self.parser_config.__class__._fields, self.parser_config)),
+            **map,
+        }
+        self.parser_config = self.Config(*(data[f] for f in self.Config._fields))
+
+    def copy(self):
+        output = object.__new__(self.__class__)
+        output.parseAction = self.parseAction[:]
+        output.parser_name = self.parser_name
+        output.token_name = self.token_name
+        output.parser_config = self.parser_config
+        output.streamlined = self.streamlined
+        output.min_length_cache = -1
+        return output
+
+    def set_parser_name(self, name):
+        """
+        Define name for this expression, makes debugging and exception messages clearer.
+
+        Example::
+
+            Word(nums).parseString("ABC")  # -> Exception: Expected W:(0123...) (at char 0), (line:1, col:1)
+            Word(nums).set_parser_name("integer").parseString("ABC")  # -> Exception: Expected integer (at char 0), (line:1, col:1)
+        """
+        self.parser_name = name
+        return self
+
+    def clearParseAction(self):
+        """
+        Add one or more parse actions to expression's list of parse actions. See `setParseAction`.
+
+        See examples in `copy`.
+        """
+        output = self.copy()
+        output.parseAction = []
+        return output
+
+    def addParseAction(self, *fns, callDuringTry=False):
+        """
+        Add one or more parse actions to expression's list of parse actions. See `setParseAction`.
+
+        See examples in `copy`.
+        """
+        output = self.copy()
+        output.parseAction += list(map(wrap_parse_action, fns))
+        output.set_config(
+            callDuringTry=self.parser_config.callDuringTry or callDuringTry
+        )
+        return output
+
+    def addCondition(self, *fns, message=None, callDuringTry=False, fatal=False):
+        """
+        Add a boolean predicate function to expression's list of parse actions. See
+        `setParseAction` for function call signatures. Unlike ``setParseAction``,
+        functions passed to ``addCondition`` need to return boolean success/fail of the condition.
+
+        Optional keyword arguments:
+        - message = define a custom message to be used in the raised exception
+        - fatal   = if True, will raise ParseFatalException to stop parsing immediately; otherwise will raise ParseException
+
+        """
+
+        def make_cond(fn):
+            def cond(token, index, string):
+                result = fn(token, index, string)
+                if not bool(result.tokens[0]):
+                    if fatal:
+                        Log.error(
+                            "fatal error",
+                            casue=ParseException(
+                                token.type, index, string, msg=message
+                            ),
+                        )
+                    raise ParseException(token.type, index, string, msg=message)
+                return token
+
+            return cond
+
+        output = self.copy()
+        for fn in fns:
+            output.parseAction.append(make_cond(wrap_parse_action(fn)))
+
+        output.set_config(
+            callDuringTry=self.parser_config.callDuringTry or callDuringTry
+        )
+        return output
+
+    def setFailAction(self, fn):
+        """Define action to perform if parsing fails at this expression.
+        Fail acton fn is a callable function that takes the arguments
+        ``fn(s, loc, expr, err)`` where:
+        - expr = the parse expression that failed
+        - loc = location where expression match was attempted and failed
+        - s = string being parsed
+        - err = the exception thrown
+        The function returns no value.  It may throw `ParseFatalException`
+        if it is desired to stop parsing immediately."""
+        self.set_config(failAction=fn)
+        return self
+
+    def is_annotated(self):
+        return self.parseAction or self.token_name or self.parser_name
+
+    def expecting(self):
+        """
+        RETURN EXPECTED CHARACTER SEQUENCE, IF ANY
+        :return:
+        """
+        return {}
+
+    def min_length(self):
+        if self.min_length_cache >= 0:
+            return self.min_length_cache
+        min_ = self._min_length()
+        if self.streamlined:
+            self.min_length_cache = min_
+        return min_
+
+    def _min_length(self):
+        return 0
+
+    @property
+    def whitespace(self):
+        return None
+
+    def parseImpl(self, string, start, doActions=True):
+        return ParseResults(self, start, start, [])
+
+    def _parse(self, string, start, doActions=True):
+        try:
+            result = self.parseImpl(string, start, doActions)
+        except Exception as cause:
+            self.parser_config.failAction and self.parser_config.failAction(
+                self, start, string, cause
+            )
+            raise
+
+        if doActions or self.parser_config.callDuringTry:
+            for fn in self.parseAction:
+                next_result = fn(result, result.start, string)
+                if next_result.end < result.end:
+                    Log.error(
+                        "parse action not allowed to roll back the end of parsing"
+                    )
+                result = next_result
+        return result
+
+    def finalize(self):
+        """
+        Return a Parser for use in parsing (optimization only)
+        :return:
+        """
+        return Parser(self)
+
+    def parseString(self, string, parseAll=False):
+        return self.finalize().parseString(string, parseAll)
+
+    def scanString(self, string, maxMatches=MAX_INT, overlap=False):
+        return (
+            self.finalize().scanString(string, maxMatches=maxMatches, overlap=overlap)
+        )
+
+    def transformString(self, string):
+        return self.finalize().transformString(string)
+
+    def searchString(self, string, maxMatches=MAX_INT):
+        return self.finalize().searchString(string, maxMatches=maxMatches)
+
+    def split(self, string, maxsplit=MAX_INT, includeSeparators=False):
+        return (
+            self
+            .finalize()
+            .split(string, maxsplit=maxsplit, includeSeparators=includeSeparators)
+        )
+
+    def replace_with(self, replacement):
+        """
+        Add parse action that replaces the token with replacement
+
+        RegEx variables are accepted:
+        \\1
+        \\g<1>
+        \\g<name>
+        """
+
+        # FIND NAMES IN replacement
+        parts = list(regex_parameters.split(replacement, includeSeparators=True))
+
+        def replacer(tokens):
+            acc = []
+            for s, n in zip(parts, parts[1:]):
+                acc.append(s)
+                acc.append(text(tokens[n]))
+            acc.append(parts[-1])
+            return "".join(acc)
+
+        return self.addParseAction(replacer)
+
+    sub = replace_with
+
     def __add__(self, other):
         """
         Implementation of + operator - returns `And`. Adding strings to a ParserElement
@@ -440,7 +506,9 @@ class ParserElement(object):
         if other is Ellipsis:
             return _PendingSkip(self)
 
-        return And([self, engine.CURRENT.normalize(other)]).streamline()
+        return And(
+            [self, whitespaces.CURRENT.normalize(other)], whitespaces.CURRENT
+        ).streamline()
 
     def __radd__(self, other):
         """
@@ -449,19 +517,19 @@ class ParserElement(object):
         if other is Ellipsis:
             return SkipTo(self)("_skipped") + self
 
-        return engine.CURRENT.normalize(other) + self
+        return whitespaces.CURRENT.normalize(other) + self
 
     def __sub__(self, other):
         """
         Implementation of - operator, returns `And` with error stop
         """
-        return self + And._ErrorStop() + engine.CURRENT.normalize(other)
+        return self + And.SyntaxErrorGuard() + whitespaces.CURRENT.normalize(other)
 
     def __rsub__(self, other):
         """
         Implementation of - operator when left operand is not a `ParserElement`
         """
-        return engine.CURRENT.normalize(other) - self
+        return whitespaces.CURRENT.normalize(other) - self
 
     def __mul__(self, other):
         """
@@ -512,7 +580,9 @@ class ParserElement(object):
                 type(other[1]),
             )
 
-        ret = Many(self, min_match=minElements, max_match=maxElements).streamline()
+        ret = Many(
+            self, whitespaces.CURRENT, min_match=minElements, max_match=maxElements
+        ).streamline()
         return ret
 
     def __rmul__(self, other):
@@ -525,37 +595,39 @@ class ParserElement(object):
         if other is Ellipsis:
             return _PendingSkip(Optional(self))
 
-        return MatchFirst([self, engine.CURRENT.normalize(other)]).streamline()
+        return MatchFirst([self, whitespaces.CURRENT.normalize(other)]).streamline()
 
     def __ror__(self, other):
         """
         Implementation of | operator when left operand is not a `ParserElement`
         """
-        return engine.CURRENT.normalize(other) | self
+        return whitespaces.CURRENT.normalize(other) | self
 
     def __xor__(self, other):
         """
         Implementation of ^ operator - returns `Or`
         """
-        return Or([self, engine.CURRENT.normalize(other)])
+        return Or([self, whitespaces.CURRENT.normalize(other)])
 
     def __rxor__(self, other):
         """
         Implementation of ^ operator when left operand is not a `ParserElement`
         """
-        return engine.CURRENT.normalize(other) ^ self
+        return whitespaces.CURRENT.normalize(other) ^ self
 
     def __and__(self, other):
         """
         Implementation of & operator - returns `Each`
         """
-        return Each([self, engine.CURRENT.normalize(other)])
+        return MatchAll(
+            [self, whitespaces.CURRENT.normalize(other)], whitespaces.CURRENT
+        )
 
     def __rand__(self, other):
         """
         Implementation of & operator when left operand is not a `ParserElement`
         """
-        return engine.CURRENT.normalize(other) & self
+        return whitespaces.CURRENT.normalize(other) & self
 
     def __invert__(self):
         """
@@ -583,8 +655,6 @@ class ParserElement(object):
         """
         output = self.copy()
         output.token_name = name
-        if not output.parser_name:
-            output.parser_name = name
         return output
 
     def suppress(self):
@@ -594,16 +664,6 @@ class ParserElement(object):
         """
         return Suppress(self)
 
-    def leaveWhitespace(self):
-        """
-        Disables the skipping of whitespace before matching the characters in the
-        `ParserElement`'s defined pattern.  This is normally only used internally by
-        the mo_parsing module, but may be needed in some whitespace-sensitive grammars.
-        """
-        with Engine(""):
-            output = self.copy()
-        return output
-
     def __str__(self):
         return self.parser_name
 
@@ -685,17 +745,12 @@ class _PendingSkip(ParserElement):
         Log.error("use of `...` expression without following SkipTo target expression")
 
 
-# export
-from mo_parsing import cache, engine, results
-
-engine.ParserElement = ParserElement
-results.ParserElement = ParserElement
-
 NO_PARSER = (
     ParserElement().set_parser_name("<nothing>")
 )  # USE THIS WHEN YOU DO NOT CARE ABOUT THE PARSER TYPE
-NO_RESULTS = ParseResults(NO_PARSER, -1, -1, [])
+NO_RESULTS = ParseResults(NO_PARSER, -1, 0, [])
 
-results.NO_PARSER = NO_PARSER
-results.NO_RESULTS = NO_RESULTS
-del results
+
+export("mo_parsing.results", ParserElement)
+export("mo_parsing.results", NO_PARSER)
+export("mo_parsing.results", NO_RESULTS)
