@@ -1,21 +1,27 @@
 # encoding: utf-8
-import re
-from collections import namedtuple
+from collections import OrderedDict
 
 from mo_dots import Null, is_null
 from mo_future import text, is_text
+from mo_imports import export, expect
 
+from mo_parsing import whitespaces
 from mo_parsing.core import ParserElement
-from mo_parsing.engine import Engine
 from mo_parsing.exceptions import (
     ParseException,
     RecursiveGrammarException,
 )
 from mo_parsing.results import ParseResults, Annotation
-from mo_parsing.utils import Log, listwrap, empty_tuple, regex_iso, append_config
+from mo_parsing.utils import (
+    Log,
+    listwrap,
+    empty_tuple,
+    regex_iso,
+    append_config,
+    regex_compile,
+)
 from mo_parsing.utils import MAX_INT, is_forward
 
-# import later
 (
     Token,
     NoMatch,
@@ -23,17 +29,29 @@ from mo_parsing.utils import MAX_INT, is_forward
     Keyword,
     Word,
     CharsNotIn,
-    _PositionToken,
     StringEnd,
     Empty,
     Char,
-) = [None] * 10
+    LookBehind,
+) = expect(
+    "Token",
+    "NoMatch",
+    "Literal",
+    "Keyword",
+    "Word",
+    "CharsNotIn",
+    "StringEnd",
+    "Empty",
+    "Char",
+    "LookBehind",
+)
 
 _get = object.__getattribute__
 
 
-class ParseElementEnhance(ParserElement):
-    """Abstract subclass of `ParserElement`, for combining and
+class ParseEnhancement(ParserElement):
+    """
+    Abstract subclass of `ParserElement`, for combining and
     post-processing parsed tokens.
     """
 
@@ -41,30 +59,31 @@ class ParseElementEnhance(ParserElement):
 
     def __init__(self, expr):
         ParserElement.__init__(self)
-        self.expr = expr = engine.CURRENT.normalize(expr)
+        self.expr = expr = whitespaces.CURRENT.normalize(expr)
         if is_forward(expr):
             expr.track(self)
 
     def copy(self):
         output = ParserElement.copy(self)
-        if self.engine is engine.CURRENT:
-            output.expr = self.expr
-        else:
-            output.expr = self.expr.copy()
+        output.expr = self.expr
         return output
 
+    def expecting(self):
+        if self.expr:
+            return OrderedDict(((k, [self]) for k, e in self.expr.expecting().items()))
+        else:
+            return {}
+
     def _min_length(self):
         return self.expr.min_length()
 
-    def parseImpl(self, string, start, doActions=True):
-        output = self.expr._parse(string, start, doActions)
-        return ParseResults(self, output.start, output.end, [output])
+    @property
+    def whitespace(self):
+        return self.expr.whitespace
 
-    def leaveWhitespace(self):
-        with Engine(""):
-            output = self.copy()
-            output.expr = self.expr.leaveWhitespace()
-            return output
+    def parseImpl(self, string, start, doActions=True):
+        result = self.expr._parse(string, start, doActions)
+        return ParseResults(self, result.start, result.end, [result])
 
     def streamline(self):
         if self.streamlined:
@@ -95,8 +114,9 @@ class ParseElementEnhance(ParserElement):
         return f"{self.__class__.__name__}:({self.expr})"
 
 
-class FollowedBy(ParseElementEnhance):
-    """Lookahead matching of the given parse expression.
+class FollowedBy(ParseEnhancement):
+    """
+    Lookahead matching of the given parse expression.
     ``FollowedBy`` does *not* advance the parsing position within
     the input string, it only verifies that the specified parse
     expression matches at the current position.  ``FollowedBy``
@@ -105,6 +125,7 @@ class FollowedBy(ParseElementEnhance):
     name.
     """
 
+    zero_length = True
     __slots__ = []
 
     def __init__(self, expr):
@@ -122,28 +143,39 @@ class FollowedBy(ParseElementEnhance):
         return "*", f"(?={self.expr.__regex__()[1]})"
 
 
-class NotAny(ParseElementEnhance):
-    """Lookahead to disallow matching with the given parse expression.
-    ``NotAny`` does *not* advance the parsing position within the
+class LookAhead(ParseEnhancement):
+    """
+    Does *not* advance the parsing position within the
     input string, it only verifies that the specified parse expression
-    does *not* match at the current position.  Also, ``NotAny`` does
-    *not* skip over leading whitespace. ``NotAny`` always returns
-    a null token list.  May be constructed using the '~' operator.
     """
 
+
+class NotAny(LookAhead):
+    """
+    Lookahead to disallow matching with the given parse expression.
+    May be constructed using the '~' operator.
+    """
+
+    zero_length = True
     __slots__ = ["regex"]
 
     def __init__(self, expr):
         super(NotAny, self).__init__(expr)
         prec, pattern = self.expr.__regex__()
         try:
-            self.regex = re.compile(f"(?!{pattern})")
-        except Exception:
-            pass
+            self.regex = regex_compile(f"(?!{pattern})")
+        except Exception as c:
+            self.regex = None
+
+    def copy(self):
+        output = ParseEnhancement.copy(self)
+        output.regex = self.regex
+        return output
 
     def parseImpl(self, string, start, doActions=True):
-        if self.regex:
-            found = self.regex.match(string, start)
+        regex = self.regex
+        if regex:
+            found = regex.match(string, start)
             if found:
                 return ParseResults(self, start, start, [])
             raise ParseException(self, start, string)
@@ -155,13 +187,18 @@ class NotAny(ParseElementEnhance):
                 return ParseResults(self, start, start, [])
 
     def streamline(self):
-        output = ParseElementEnhance.streamline(self)
+        output = ParseEnhancement.streamline(self)
+        if isinstance(output, Empty):
+            return NoMatch()
         if isinstance(output.expr, NoMatch):
             return Empty()
         if isinstance(output.expr, Empty):
             return NoMatch()
         return output
 
+    def expecting(self):
+        return {}
+
     def min_length(self):
         return 0
 
@@ -174,11 +211,21 @@ class NotAny(ParseElementEnhance):
         return "~{" + str(self.expr) + "}"
 
 
-class Many(ParseElementEnhance):
+class Many(ParseEnhancement):
     __slots__ = []
-    Config = append_config(ParseElementEnhance, "min_match", "max_match", "end")
+    Config = append_config(
+        ParseEnhancement, "whitespace", "min_match", "max_match", "end"
+    )
 
-    def __init__(self, expr, stopOn=None, min_match=0, max_match=MAX_INT, exact=None):
+    def __init__(
+        self,
+        expr,
+        whitespace=None,
+        stopOn=None,
+        min_match=0,
+        max_match=MAX_INT,
+        exact=None,
+    ):
         """
         MATCH expr SOME NUMBER OF TIMES (OR UNTIL stopOn IS REACHED
         :param expr: THE EXPRESSION TO MATCH
@@ -186,18 +233,22 @@ class Many(ParseElementEnhance):
         :param min_match: MINIMUM MATCHES REQUIRED FOR SUCCESS (-1 IS INVALID)
         :param max_match: MAXIMUM MATCH REQUIRED FOR SUCCESS (-1 IS INVALID)
         """
-        ParseElementEnhance.__init__(self, expr)
+        ParseEnhancement.__init__(self, expr)
         if exact is not None:
             min_match = exact
             max_match = exact
 
-        self.set_config(min_match=min_match, max_match=max_match)
+        self.set_config(
+            whitespace=whitespace or whitespaces.CURRENT,
+            min_match=min_match,
+            max_match=max_match,
+        )
         self.stopOn(stopOn)
 
     def stopOn(self, ender):
         if ender:
-            end = self.engine.normalize(ender)
-            self.set_config(end=re.compile(end.__regex__()[1]))
+            end = self.parser_config.whitespace.normalize(ender)
+            self.set_config(end=regex_compile(end.__regex__()[1]))
         return self
 
     def _min_length(self):
@@ -205,33 +256,43 @@ class Many(ParseElementEnhance):
             return 0
         return self.expr.min_length()
 
+    @property
+    def whitespace(self):
+        return self.parser_config.whitespace
+
     def parseImpl(self, string, start, doActions=True):
         acc = []
-        end = start
+        end = index = start
         max = self.parser_config.max_match
         stopper = self.parser_config.end
         count = 0
         try:
             while end < len(string) and count < max:
+                if end > index:
+                    if isinstance(self.expr, LookBehind):
+                        index = end
+                    else:
+                        index = self.parser_config.whitespace.skip(string, end)
                 if stopper:
-                    end = self.engine.skip(string, end)
-                    if stopper.match(string, end):
+                    if stopper.match(string, index):
                         if self.parser_config.min_match <= count:
                             break
                         else:
                             raise ParseException(
                                 self, end, string, msg="found stopper too soon"
                             )
-                tokens = self.expr._parse(string, end, doActions)
-                end = tokens.end
-                if tokens:
-                    acc.append(tokens)
+                result = self.expr._parse(string, index, doActions)
+                end = result.end
+                if result:
+                    acc.append(result)
                     count += 1
-        except ParseException:
+        except ParseException as pe:
             if self.parser_config.min_match <= count <= max:
                 pass
             else:
-                ParseException(self, start, string, msg="Not correct amount of matches")
+                raise ParseException(
+                    self, start, string, msg="Not correct amount of matches"
+                )
         if count:
             if (
                 count < self.parser_config.min_match
@@ -262,7 +323,10 @@ class Many(ParseElementEnhance):
     def streamline(self):
         if self.streamlined:
             return self
-        expr = self.expr.streamline()
+        try:
+            expr = self.expr.streamline()
+        except Exception as e:
+            print(e)
         if (
             self.parser_config.min_match == self.parser_config.max_match
             and not self.is_annotated()
@@ -324,7 +388,7 @@ class Many(ParseElementEnhance):
                     " expressions"
                 )
 
-        return ParseElementEnhance.__call__(self, name)
+        return ParseEnhancement.__call__(self, name)
 
     def __str__(self):
         if self.parser_name:
@@ -344,15 +408,13 @@ class OneOrMore(Many):
 
     __slots__ = []
 
-    def __init__(self, expr, stopOn=None):
-        Many.__init__(self, expr, stopOn, min_match=1, max_match=MAX_INT)
-        self.set_config(lock_engine=self.expr.parser_config.lock_engine,)
-        self.engine = self.expr.engine
+    def __init__(self, expr, whitespace=None, stopOn=None):
+        Many.__init__(self, expr, whitespace, stopOn, min_match=1, max_match=MAX_INT)
 
     def __str__(self):
         if self.parser_name:
             return self.parser_name
-        return "{" + text(self.expr) + "}..."
+        return "{" + text(self.expr) + "}+"
 
 
 class ZeroOrMore(Many):
@@ -369,12 +431,10 @@ class ZeroOrMore(Many):
 
     __slots__ = []
 
-    def __init__(self, expr, stopOn=None):
+    def __init__(self, expr, whitespace=None, stopOn=None):
         super(ZeroOrMore, self).__init__(
-            expr, stopOn=stopOn, min_match=0, max_match=MAX_INT
+            expr, whitespace, stopOn=stopOn, min_match=0, max_match=MAX_INT
         )
-        self.set_config(lock_engine=self.expr.parser_config.lock_engine)
-        self.engine = self.expr.engine
 
     def parseImpl(self, string, start, doActions=True):
         try:
@@ -386,7 +446,7 @@ class ZeroOrMore(Many):
         if self.parser_name:
             return self.parser_name
 
-        return "[" + text(self.expr) + "]..."
+        return "(" + text(self.expr) + ")*"
 
 
 class Optional(Many):
@@ -400,14 +460,14 @@ class Optional(Many):
     __slots__ = []
     Config = append_config(Many, "defaultValue")
 
-    def __init__(self, expr, default=None):
-        Many.__init__(self, expr, stopOn=None, min_match=0, max_match=1)
+    def __init__(self, expr, whitespace=None, default=None):
+        Many.__init__(self, expr, whitespace, stopOn=None, min_match=0, max_match=1)
         self.set_config(defaultValue=listwrap(default))
 
     def parseImpl(self, string, start, doActions=True):
         try:
-            tokens = self.expr._parse(string, start, doActions)
-            return ParseResults(self, tokens.start, tokens.end, [tokens])
+            results = self.expr._parse(string, start, doActions)
+            return ParseResults(self, results.start, results.end, [results])
         except ParseException:
             return ParseResults(self, start, start, self.parser_config.defaultValue)
 
@@ -418,13 +478,13 @@ class Optional(Many):
         return "[" + text(self.expr) + "]"
 
 
-class SkipTo(ParseElementEnhance):
+class SkipTo(ParseEnhancement):
     """Token for skipping over all undefined text until the matched expression is found."""
 
     __slots__ = []
-    Config = append_config(ParseElementEnhance, "include", "fail", "ignore")
+    Config = append_config(ParseEnhancement, "include", "fail", "ignore", "whitespace")
 
-    def __init__(self, expr, include=False, ignore=None, failOn=None):
+    def __init__(self, expr, include=False, ignore=None, failOn=None, engine_=None):
         """
         :param expr: target expression marking the end of the data to be skipped
         :param include: if True, the target expression is also parsed
@@ -435,9 +495,13 @@ class SkipTo(ParseElementEnhance):
           included in the skipped test; if found before the target expression is found,
           the SkipTo is not a match
         """
-        ParseElementEnhance.__init__(self, expr)
+        ParseEnhancement.__init__(self, expr)
+
         self.set_config(
-            include=include, fail=engine.CURRENT.normalize(failOn), ignore=ignore
+            include=include,
+            fail=whitespaces.CURRENT.normalize(failOn),
+            ignore=ignore,
+            whitespace=engine_ or whitespaces.CURRENT,
         )
         self.parser_name = str(self)
 
@@ -451,40 +515,42 @@ class SkipTo(ParseElementEnhance):
 
         loc = start
         while loc <= instrlen:
+            skip_end = loc
+            loc = before_end = self.parser_config.whitespace.skip(string, loc)
             if fail:
                 # break if failOn expression matches
-                if fail.canParseNext(string, loc):
-                    before_end = loc
+                try:
+                    fail._parse(string, loc)
                     break
+                except:
+                    pass
 
             if ignore:
                 # advance past ignore expressions
                 while 1:
                     try:
-                        loc = ignore.tryParse(string, loc)
-                        if loc == None:
-                            Log.error("")
+                        loc = ignore._parse(string, loc).end
+                        skip_end = loc
+                        loc = before_end = self.parser_config.whitespace.skip(
+                            string, loc
+                        )
                     except ParseException:
                         break
             try:
-                before_end = loc
                 loc = self.expr._parse(string, loc, doActions=False).end
-                if loc == None:
-                    Log.error("")
             except ParseException:
                 # no match, advance loc in string
                 loc += 1
             else:
                 # matched skipto expr, done
                 break
-
         else:
             # ran off the end of the input string without matching skipto expr, fail
             raise ParseException(self, start, string)
 
         # build up return values
         end = loc
-        skiptext = string[start:before_end]
+        skiptext = string[start:skip_end]
         skip_result = []
         if skiptext:
             skip_result.append(skiptext)
@@ -525,7 +591,7 @@ class Forward(ParserElement):
     parser created using ``Forward``.
     """
 
-    __slots__ = ["expr", "used_by", "_str"]
+    __slots__ = ["expr", "used_by", "_str", "_reg", "_eng"]
 
     def __init__(self, expr=Null):
         ParserElement.__init__(self)
@@ -533,13 +599,17 @@ class Forward(ParserElement):
         self.used_by = []
 
         self._str = None  # avoid recursion
+        self._reg = None  # avoid recursion
+        self._eng = False
         if expr:
-            self << engine.CURRENT.normalize(expr)
+            self << whitespaces.CURRENT.normalize(expr)
 
     def copy(self):
         output = ParserElement.copy(self)
         output.expr = self
         output._str = None
+        output._reg = None
+        output._eng = False
 
         output.used_by = []
         return output
@@ -558,7 +628,10 @@ class Forward(ParserElement):
 
         while is_forward(other):
             other = other.expr
-        self.expr = engine.CURRENT.normalize(other).streamline()
+        norm = whitespaces.CURRENT.normalize(other)
+        if norm is None:
+            Log.error("problem")
+        self.expr = norm.streamline()
         self.checkRecursion()
         return self
 
@@ -566,9 +639,10 @@ class Forward(ParserElement):
         if not self.expr:
             Log.error("not allowed")
         self.expr = self.expr.addParseAction(action)
+        return self
 
     def leaveWhitespace(self):
-        with Engine(""):
+        with whitespaces.NO_WHITESPACE:
             output = self.copy()
             output.expr = self.expr.leaveWhitespace()
             return output
@@ -588,14 +662,29 @@ class Forward(ParserElement):
             self.expr.checkRecursion(seen + (self,))
 
     def min_length(self):
-        if self.min_cache is None and self.expr:
-            self.min_cache = 0  # BREAK CYCLE
+        if self.min_length_cache is None and self.expr:
+            self.min_length_cache = 0  # BREAK CYCLE
             try:
                 return self.expr.min_length()
             finally:
-                self.min_cache = None
+                self.min_length_cache = None
         return 0
 
+    @property
+    def whitespace(self):
+        try:
+            if self._eng:
+                return None
+        except Exception as cause:
+            Log.error("", cause=cause)
+
+        # Avoid infinite recursion by setting a temporary
+        self._eng = True
+        try:
+            return self.expr.whitespace
+        finally:
+            self._eng = False
+
     def parseImpl(self, string, loc, doActions=True):
         try:
             result = self.expr._parse(string, loc, doActions)
@@ -608,6 +697,16 @@ class Forward(ParserElement):
                 )
             raise cause
 
+    def __regex__(self):
+        if self._reg or not self.expr:
+            return None
+
+        try:
+            self._reg = True
+            return self.expr.__regex__()
+        finally:
+            self._reg = None
+
     def __str__(self):
         if self.parser_name:
             return self.parser_name
@@ -629,7 +728,7 @@ class Forward(ParserElement):
         return output
 
 
-class TokenConverter(ParseElementEnhance):
+class TokenConverter(ParseEnhancement):
     """
     Abstract subclass of `ParseExpression`, for converting parsed results.
     """
@@ -645,24 +744,43 @@ class Combine(TokenConverter):
     Converter to concatenate all matching tokens to a single string.
     """
 
-    __slots__ = []
     Config = append_config(TokenConverter, "separator")
 
     def __init__(self, expr, separator=""):
         super(Combine, self).__init__(expr.streamline())
         self.set_config(separator=separator)
-        self.parseAction.append(_combine)
-        self.streamlined = True
 
+    def parseImpl(self, string, start, doActions=True):
+        result = self.expr.parseImpl(string, start, doActions=doActions)
+        output = ParseResults(
+            self,
+            start,
+            result.end,
+            [result.asString(sep=self.parser_config.separator)],
+        )
+        return output
 
-def _combine(tokens, start, string):
-    output = ParseResults(
-        tokens.type,
-        tokens.start,
-        tokens.end,
-        [tokens.asString(sep=tokens.type.parser_config.separator)],
-    )
-    return output
+    def streamline(self):
+        if self.streamlined:
+            return self
+
+        expr = self.expr.streamline()
+        if expr is self.expr:
+            self.streamlined = True
+            return self
+        return Combine(expr, self.parser_config.separator)
+
+    def expecting(self):
+        return OrderedDict((k, [self]) for k in self.expr.expecting().keys())
+
+    def min_length(self):
+        return self.expr.min_length()
+
+    def __regex__(self):
+        return self.expr.__regex__()
+
+    def __str__(self):
+        return text(self.expr)
 
 
 class Group(TokenConverter):
@@ -674,7 +792,7 @@ class Group(TokenConverter):
 
     def __init__(self, expr):
         ParserElement.__init__(self)
-        self.expr = self.engine.normalize(expr)
+        self.expr = whitespaces.CURRENT.normalize(expr)
 
     def is_annotated(self):
         return True
@@ -755,8 +873,9 @@ def _suppress_post_parse(tokens, start, string):
     return ParseResults(tokens.type, tokens.start, tokens.end, [])
 
 
-class PrecededBy(ParseElementEnhance):
-    """Lookbehind matching of the given parse expression.
+class PrecededBy(LookAhead):
+    """
+    Lookbehind matching of the given parse expression.
     ``PrecededBy`` does not advance the parsing position within the
     input string, it only verifies that the specified parse expression
     matches prior to the current position.  ``PrecededBy`` always
@@ -777,18 +896,19 @@ class PrecededBy(ParseElementEnhance):
     the current parse position for a lookbehind match.
     """
 
+    zero_length = True
     __slots__ = []
-    Config = append_config(ParseElementEnhance, "retreat", "exact")
+    Config = append_config(ParseEnhancement, "retreat", "exact")
 
     def __init__(self, expr, retreat=None):
         super(PrecededBy, self).__init__(expr)
-        expr = self.expr = self.expr.leaveWhitespace()
+        expr = self.expr
 
         if isinstance(expr, (Literal, Keyword, Char)):
             self.set_config(retreat=expr.min_length(), exact=True)
         elif isinstance(expr, (Word, CharsNotIn)):
             self.set_config(retreat=expr.min_length(), exact=False)
-        elif isinstance(expr, _PositionToken):
+        elif expr.__class__.zero_length:
             self.set_config(retreat=0, exact=True)
         else:
             self.set_config(retreat=expr.min_length(), exact=False)
@@ -805,15 +925,14 @@ class PrecededBy(ParseElementEnhance):
             instring_slice = string[:start]
             last_cause = ParseException(self, start, string)
 
-            with self.engine.backup():
-                for offset in range(self.parser_config.retreat, start + 1):
-                    try:
-                        ret = test_expr._parse(instring_slice, start - offset)
-                        break
-                    except ParseException as cause:
-                        last_cause = cause
-                else:
-                    raise last_cause
+            for offset in range(self.parser_config.retreat, start + 1):
+                try:
+                    ret = test_expr._parse(instring_slice, start - offset)
+                    break
+                except ParseException as cause:
+                    last_cause = cause
+            else:
+                raise last_cause
         # return empty list of tokens, but preserve any defined results names
 
         ret.__class__ = Annotation
@@ -825,18 +944,19 @@ class PrecededBy(ParseElementEnhance):
         raise NotImplemented()
 
 
-# export
-from mo_parsing import core, engine, results
+export("mo_parsing.core", SkipTo)
+export("mo_parsing.core", Many)
+export("mo_parsing.core", ZeroOrMore)
+export("mo_parsing.core", OneOrMore)
+export("mo_parsing.core", Optional)
+export("mo_parsing.core", NotAny)
+export("mo_parsing.core", Suppress)
+export("mo_parsing.core", Group)
 
-core.SkipTo = SkipTo
-core.Many = Many
-core.ZeroOrMore = ZeroOrMore
-core.OneOrMore = OneOrMore
-core.Optional = Optional
-core.NotAny = NotAny
-core.Suppress = Suppress
-core.Group = Group
+export("mo_parsing.results", Group)
+export("mo_parsing.results", Dict)
+export("mo_parsing.results", Suppress)
 
-results.Group = Group
-results.Dict = Dict
-results.Suppress = Suppress
+export("mo_parsing.whitespaces", Empty)
+
+export("mo_parsing.utils", Many)
