@@ -1,25 +1,32 @@
 # encoding: utf-8
+import json
+from collections import OrderedDict
 from operator import itemgetter
 
 from mo_future import Iterable, text, generator_types
+from mo_imports import export
 
+from mo_parsing import whitespaces
 from mo_parsing.core import ParserElement, _PendingSkip
-from mo_parsing.engine import Engine
 from mo_parsing.enhancement import Optional, SkipTo, Many
 from mo_parsing.exceptions import (
     ParseException,
     ParseSyntaxException,
 )
 from mo_parsing.results import ParseResults
-from mo_parsing.tokens import Empty
+from mo_parsing.tokens import Empty, LookBehind
 from mo_parsing.utils import (
     empty_tuple,
     is_forward,
     regex_iso,
     Log,
     append_config,
-    stack_depth,
+    regex_caseless,
+    regex_compile,
 )
+from mo_parsing.whitespaces import Whitespace
+
+LOOKUP_COST = 4
 
 
 class ParseExpression(ParserElement):
@@ -39,17 +46,34 @@ class ParseExpression(ParserElement):
         else:
             exprs = [exprs]
 
-        self.exprs = [engine.CURRENT.normalize(e) for e in exprs]
+        self.exprs = [whitespaces.CURRENT.normalize(e) for e in exprs]
         for e in self.exprs:
             if is_forward(e):
                 e.track(self)
 
+    def expecting(self):
+        output = OrderedDict()
+        if not self.is_annotated():
+            for e in self.exprs:
+                expect = e.expecting()
+                if not expect:
+                    # NOT SURE WHAT THIS IS EXPECTING, BAIL
+                    return {}
+                for k, ee in expect.items():
+                    output.setdefault(k, []).extend(ee)
+        else:
+            for e in self.exprs:
+                expect = e.expecting()
+                if not expect:
+                    # NOT SURE WHAT THIS IS EXPECTING, BAIL
+                    return {}
+                for k, _ in expect.items():
+                    output[k] = [self]
+        return output
+
     def copy(self):
         output = ParserElement.copy(self)
-        if self.engine is engine.CURRENT:
-            output.exprs = self.exprs
-        else:
-            output.exprs = [e.copy() for e in self.exprs]
+        output.exprs = self.exprs
         return output
 
     def append(self, other):
@@ -59,7 +83,7 @@ class ParseExpression(ParserElement):
     def leaveWhitespace(self):
         """Extends ``leaveWhitespace`` defined in base class, and also invokes ``leaveWhitespace`` on
         all contained expressions."""
-        with Engine(""):
+        with whitespaces.NO_WHITESPACE:
             output = self.copy()
             output.exprs = [e.leaveWhitespace() for e in self.exprs]
             return output
@@ -77,12 +101,18 @@ class ParseExpression(ParserElement):
 
         acc = []
         same = True
+        clazz = self.__class__
+        if clazz == Or:
+            clazz = (
+                Or,
+                MatchFirst,
+            )  # TODO: not correct, but allows merging of the two to a single longer list
         for e in self.exprs:
             f = e.streamline()
             same = same and f is e
             if f.is_annotated():
                 acc.append(f)
-            elif isinstance(f, self.__class__):
+            elif isinstance(f, clazz):
                 same = False
                 acc.extend(f.exprs)
             else:
@@ -99,10 +129,6 @@ class ParseExpression(ParserElement):
     def __call__(self, name):
         if not name:
             return self
-        # for e in self.exprs:
-        #     if isinstance(e, ParserElement) and e.token_name:
-        #         Log.error("token name is already set in child, use Group() to clarify")
-
         return ParserElement.__call__(self, name)
 
 
@@ -113,19 +139,18 @@ class And(ParseExpression):
     May be constructed using the ``'+'`` operator.
     May also be constructed using the ``'-'`` operator, which will
     suppress backtracking.
-
     """
 
     __slots__ = []
+    Config = append_config(ParseExpression, "whitespace")
 
-    class _ErrorStop(Empty):
+    class SyntaxErrorGuard(Empty):
         def __init__(self, *args, **kwargs):
-            with Engine() as engine:
-                engine.set_whitespace("")
-                super(And._ErrorStop, self).__init__(*args, **kwargs)
+            with Whitespace(""):
+                super(And.SyntaxErrorGuard, self).__init__(*args, **kwargs)
                 self.parser_name = "-"
 
-    def __init__(self, exprs):
+    def __init__(self, exprs, whitespace=None):
         if exprs and Ellipsis in exprs:
             tmp = []
             for i, expr in enumerate(exprs):
@@ -141,6 +166,7 @@ class And(ParseExpression):
                     tmp.append(expr)
             exprs[:] = tmp
         super(And, self).__init__(exprs)
+        self.set_config(whitespace=whitespace or whitespaces.CURRENT)
 
     def streamline(self):
         if self.streamlined:
@@ -180,7 +206,10 @@ class And(ParseExpression):
             same = same and f is e
             if f.is_annotated():
                 acc.append(f)
-            elif isinstance(f, self.__class__):
+            elif (
+                isinstance(f, And)
+                and f.parser_config.whitespace is self.parser_config.whitespace
+            ):
                 same = False
                 acc.extend(f.exprs)
             else:
@@ -195,25 +224,49 @@ class And(ParseExpression):
         output.streamlined = True
         return output
 
+    def expecting(self):
+        if not self.exprs:
+            return {}
+
+        acc = OrderedDict()
+        for e in self.exprs:
+            expect = e.expecting()
+            if not expect:
+                return {}
+            for k in expect.keys():
+                acc[k] = [self]
+            if e.min_length():
+                break
+        return acc
+
     def _min_length(self):
         return sum(e.min_length() for e in self.exprs)
 
+    @property
+    def whitespace(self):
+        return self.parser_config.whitespace
+
     def parseImpl(self, string, start, doActions=True):
         # pass False as last arg to _parse for first element, since we already
         # pre-parsed the string as part of our And pre-parsing
-        encountered_error_stop = False
-        end = start
+        encountered_syntax_error = False
+        end = index = start
         acc = []
-        for expr in self.exprs:
-            if isinstance(expr, And._ErrorStop):
-                encountered_error_stop = True
+        for i, expr in enumerate(self.exprs):
+            if end > index:
+                if isinstance(expr, LookBehind):
+                    index = end
+                else:
+                    index = self.parser_config.whitespace.skip(string, end)
+            if isinstance(expr, And.SyntaxErrorGuard):
+                encountered_syntax_error = True
                 continue
             try:
-                exprtokens = expr._parse(string, end, doActions)
-                end = exprtokens.end
-                acc.append(exprtokens)
+                result = expr._parse(string, index, doActions)
+                end = result.end
+                acc.append(result)
             except ParseException as pe:
-                if encountered_error_stop:
+                if encountered_syntax_error:
                     raise ParseSyntaxException(pe.expr, pe.loc, pe.string)
                 else:
                     raise pe
@@ -224,7 +277,9 @@ class And(ParseExpression):
         if other is Ellipsis:
             return _PendingSkip(self)
 
-        return And([self, engine.CURRENT.normalize(other)]).streamline()
+        return And(
+            [self, whitespaces.CURRENT.normalize(other)], whitespaces.CURRENT
+        ).streamline()
 
     def checkRecursion(self, seen=empty_tuple):
         subRecCheckList = seen + (self,)
@@ -240,33 +295,95 @@ class And(ParseExpression):
         if self.parser_name:
             return self.parser_name
 
-        return "{" + " + ".join(text(e) for e in self.exprs) + "}"
+        subs = [text(e) for e in self.exprs]
+        if all(len(s) == 1 for s in subs):
+            return "".join(subs)
+        else:
+            return " + ".join(
+                "{" + text(e) + "}" if isinstance(e, MatchFirst) else text(e)
+                for e in self.exprs
+            )
 
 
 class Or(ParseExpression):
-    """Requires that at least one `ParseExpression` is found. If
+    """
+    Requires that at least one `ParseExpression` is found. If
     two expressions match, the expression that matches the longest
     string will be used. May be constructed using the ``'^'``
     operator.
     """
 
-    __slots__ = []
+    __slots__ = ["alternate"]
 
     def __init__(self, exprs):
-        super(Or, self).__init__(exprs)
+        ParseExpression.__init__(self, exprs)
+        self.alternate = self.exprs
+
+    def copy(self):
+        output = ParseExpression.copy(self)
+        output.alternate = self.alternate
+        return output
 
     def _min_length(self):
         return min(e.min_length() for e in self.exprs)
 
+    def streamline(self):
+        if self.streamlined:
+            return self
+
+        output = ParseExpression.streamline(self)
+
+        if not isinstance(output, ParseExpression):
+            return output
+        if not output.is_annotated():
+            if len(output.exprs) == 0:
+                output = Empty()
+            if len(output.exprs) == 1:
+                output = output.exprs[0]
+                if not isinstance(output, ParseExpression):
+                    return output
+
+        output.alternate = faster(output.exprs)
+
+        output.streamlined = True
+        output.checkRecursion()
+        return output
+
+    @property
+    def whitespace(self):
+        return [e.whitespace for e in self.exprs]
+
     def parseImpl(self, string, start, doActions=True):
         causes = []
         matches = []
-        for e in self.exprs:
-            try:
-                end = e.tryParse(string, start)
-                matches.append((end, e))
-            except ParseException as err:
-                causes.append(err)
+
+        for e in self.alternate:
+            if isinstance(e, Fast):
+                for ee in e.get_short_list(string, start):
+                    try:
+                        end = ee._parse(string, start).end
+                        matches.append((end, ee))
+                    except ParseException as err:
+                        causes.append(err)
+            else:
+                try:
+                    end = e._parse(string, start).end
+                    matches.append((end, e))
+                except ParseException as err:
+                    causes.append(err)
+
+        if not matches:
+            raise ParseException(
+                self,
+                start,
+                string,
+                msg="no defined alternatives to match",
+                cause=causes,
+            )
+        if len(matches) == 1:
+            _, expr = matches[0]
+            result = expr._parse(string, start, doActions)
+            return ParseResults(self, result.start, result.end, [result])
 
         if matches:
             # re-evaluate all matches in descending order of length of match, in case attached actions
@@ -276,39 +393,33 @@ class Or(ParseExpression):
             if not doActions:
                 # no further conditions or parse actions to change the selection of
                 # alternative, so the first match will be the best match
-                _, best_expr = matches[0]
-                best_results = best_expr._parse(string, start, doActions)
-                return ParseResults(
-                    self, best_results.start, best_results.end, [best_results]
-                )
+                _, expr = matches[0]
+                result = expr._parse(string, start, doActions)
+                return ParseResults(self, result.start, result.end, [result])
 
             longest = -1, None
-            for loc, expr1 in matches:
+            for loc, expr in matches:
                 if loc <= longest[0]:
                     # already have a longer match than this one will deliver, we are done
                     return longest
 
                 try:
-                    toks = expr1._parse(string, start, doActions)
+                    result = expr._parse(string, start, doActions)
                 except ParseException as err:
                     causes.append(err)
                 else:
-                    if toks.end >= loc:
-                        return ParseResults(self, toks.start, toks.end, [toks])
+                    if result.end >= loc:
+                        return ParseResults(self, result.start, result.end, [result])
                     # didn't match as much as before
-                    elif toks.end > longest[0]:
+                    elif result.end > longest[0]:
                         longest = (
-                            toks.end,
-                            ParseResults(self, toks.start, toks.end, [toks]),
+                            result.end,
+                            ParseResults(self, result.start, result.end, [result]),
                         )
 
             if longest != (-1, None):
                 return longest
 
-        raise ParseException(
-            self, start, string, msg="no defined alternatives to match", cause=causes
-        )
-
     def checkRecursion(self, seen=empty_tuple):
         seen_more = seen + (self,)
         for e in self.exprs:
@@ -332,15 +443,22 @@ class Or(ParseExpression):
 
 
 class MatchFirst(ParseExpression):
-    """Requires that at least one `ParseExpression` is found. If
+    """
+    Requires that at least one `ParseExpression` is found. If
     two expressions match, the first one listed is the one that will
-    match. May be constructed using the ``'|'`` operator.
+    match. May be constructed using the `|` operator.
     """
 
-    __slots__ = []
+    __slots__ = ["alternate"]
 
     def __init__(self, exprs):
-        super(MatchFirst, self).__init__(exprs)
+        ParseExpression.__init__(self, exprs)
+        self.alternate = self.exprs
+
+    def copy(self):
+        output = ParseExpression.copy(self)
+        output.alternate = self.alternate
+        return output
 
     def _min_length(self):
         if self.exprs:
@@ -349,13 +467,17 @@ class MatchFirst(ParseExpression):
             Log.warning("expecting streamline")
             return 0
 
+    @property
+    def whitespace(self):
+        return [e.whitespace for e in self.exprs]
+
     def parseImpl(self, string, start, doActions=True):
         causes = []
 
-        for e in self.exprs:
+        for e in self.alternate:
             try:
-                ret = e._parse(string, start, doActions)
-                return ParseResults(self, ret.start, ret.end, [ret])
+                result = e._parse(string, start, doActions)
+                return ParseResults(self, result.start, result.end, [result])
             except ParseException as cause:
                 causes.append(cause)
 
@@ -371,9 +493,11 @@ class MatchFirst(ParseExpression):
             return output
         if not output.is_annotated():
             if len(output.exprs) == 0:
-                output = Empty()
+                return Empty()
             if len(output.exprs) == 1:
-                output = output.exprs[0]
+                return output.exprs[0]
+
+        output.alternate = faster(output.exprs)
 
         output.streamlined = True
         output.checkRecursion()
@@ -388,10 +512,10 @@ class MatchFirst(ParseExpression):
         if other is Ellipsis:
             return _PendingSkip(Optional(self))
 
-        return MatchFirst([self, engine.CURRENT.normalize(other)]).streamline()
+        return MatchFirst([self, whitespaces.CURRENT.normalize(other)]).streamline()
 
     def __ror__(self, other):
-        return engine.CURRENT.normalize(other) | self
+        return whitespaces.CURRENT.normalize(other) | self
 
     def __regex__(self):
         return (
@@ -410,7 +534,145 @@ class MatchFirst(ParseExpression):
         return " | ".join("{" + text(e) + "}" for e in self.exprs)
 
 
-class Each(ParseExpression):
+def faster(exprs):
+    """
+    BUILD A LOOKUP ARRAY TO MATCH ANY OF THE GIVEN exprs
+    PERFORMS A REGEX, AND USES THE lower() CHARACTERS TO JUMP TO A SHORTLIST OF exprs THAT CAN MATCH
+    :param exprs:
+    :return: LESS EXPRESSIONS
+    """
+
+    if len(exprs) < LOOKUP_COST:
+        return exprs
+
+    alternating = []
+    # SOME NUMBER OF CONSTANT PATTERNS
+    acc = []
+    out = []
+    has_expecting = True
+    for o in exprs:
+        p = o.expecting()
+        if has_expecting:
+            if p:
+                acc.append(p)
+                out.append(o)
+            else:
+                try:
+                    e = Fast(acc)
+                    alternating.append(e)
+                except Exception as c:
+                    alternating.extend(out)
+                acc = []
+                out = []
+                alternating.append(o)
+                has_expecting = False
+        elif p:
+            acc = [p]
+            out = [o]
+            has_expecting = True
+        else:
+            alternating.append(o)
+
+    if has_expecting:
+        try:
+            e = Fast(acc)
+            alternating.append(e)
+        except Exception as cause:
+            alternating.extend(out)
+    return alternating
+
+
+def _distinct(a, b):
+    """
+    ASSUME a != b
+    RETURN MINIMUM length SO THAT a[:length] != b[:length]
+    """
+    ii = 1
+    for aa, bb in zip(a, b):
+        if aa != bb:
+            return ii
+        ii += 1
+    return ii
+
+
+class Fast(ParserElement):
+    __slots__ = ["lookup", "regex", "all_keys"]
+
+    def __init__(self, maps):
+        ParserElement.__init__(self)
+
+        all_keys = set()
+        lookup = OrderedDict()
+        for m in maps:
+            for k, ee in m.items():
+                all_keys.add(k)
+                lookup.setdefault(k, []).extend(ee)
+
+        # patterns must be mutually exclusive to work
+        items = list(lookup.items())
+        if len(maps) - max(len(v) for k, v in items) < LOOKUP_COST:
+            Log.error("not useful")
+
+        compact = []
+        for k, e in items:
+            min_k = k
+            # FIND SHORTEST PREFIX
+            for kk, ee in items:
+                if ee and min_k.startswith(kk):
+                    min_k = kk
+            # COLLECT
+            acc = []
+            for kk, ee in items:
+                if kk.startswith(min_k):
+                    acc.extend(ee)
+                    ee.clear()
+            if acc:
+                compact.append((min_k, acc))
+        if len(maps) - max(len(v) for k, v in compact) < LOOKUP_COST:
+            Log.error("not useful")
+
+        # patterns can be shortened so far as they remain exclusive
+        shorter = [
+            (k[:min_length], e)
+            for k, e in sorted(compact, key=lambda p: p[0])
+            for min_length in [max(_distinct(k, kk) for kk, _ in compact if kk != k)]
+        ]
+
+        self.lookup = {k: e for k, e in shorter}
+        self.regex = regex_compile("|".join(regex_caseless(k) for k, _ in shorter))
+        self.all_keys = list(sorted(all_keys))
+
+    def get_short_list(self, string, start):
+        """
+        USE THE LOOKUP FEATURE TO FIND THE FEW ParserElements THAT CAN MATCH
+        """
+        found = self.regex.match(string, start)
+        if found:
+            index = found.group(0).lower()
+            return self.lookup[index]
+        return []
+
+    def parseImpl(self, string, start, doActions=True):
+        found = self.regex.match(string, start)
+        if found:
+            index = found.group(0).lower()
+            exprs = self.lookup[index]
+
+            causes = []
+            for e in exprs:
+                try:
+                    return e._parse(string, start, doActions)
+                except ParseException as cause:
+                    causes.append(cause)
+
+            raise ParseException(self, start, string, cause=causes)
+        else:
+            raise ParseException(
+                self, start, string, "expecting one of " + json.dumps(self.all_keys)
+            )
+
+
+class MatchAll(ParseExpression):
     """
     Requires all given `ParseExpression` s to be found, but in
     any order. Expressions may be separated by whitespace.
@@ -419,15 +681,16 @@ class Each(ParseExpression):
     """
 
     __slots__ = []
-    Config = append_config(ParseExpression, "min_match", "max_match")
+    Config = append_config(ParseExpression, "min_match", "max_match", "whitespace")
 
-    def __init__(self, exprs):
+    def __init__(self, exprs, whitespace):
         """
         :param exprs: The expressions to be matched
         :param mins: list of integers indincating any minimums
         """
-        super(Each, self).__init__(exprs)
+        super(MatchAll, self).__init__(exprs)
         self.set_config(
+            whitespace=whitespace,
             min_match=[
                 e.parser_config.min_match if isinstance(e, Many) else 1 for e in exprs
             ],
@@ -439,12 +702,16 @@ class Each(ParseExpression):
     def streamline(self):
         if self.streamlined:
             return self
-        return super(Each, self).streamline()
+        return super(MatchAll, self).streamline()
 
     def _min_length(self):
         # TODO: MAY BE TOO CONSERVATIVE, WE MAY BE ABLE TO PROVE self CAN CONSUME A CHARACTER
         return min(e.min_length() for e in self.exprs)
 
+    @property
+    def whitespace(self):
+        return [e.whitespace for e in self.exprs]
+
     def parseImpl(self, string, start, doActions=True):
         end = start
         matchOrder = []
@@ -456,10 +723,10 @@ class Each(ParseExpression):
         while todo:
             for i, (c, (e, mi, ma)) in enumerate(zip(count, todo)):
                 try:
-                    loc = e.tryParse(string, end)
+                    loc = e._parse(string, end).end
                     if loc == end:
                         continue
-                    end = loc
+                    end = self.parser_config.whitespace.skip(string, loc)
                     c2 = count[i] = c + 1
                     if c2 >= ma:
                         del todo[i]
@@ -474,8 +741,9 @@ class Each(ParseExpression):
         for c, (e, mi, ma) in zip(count, todo):
             if c < mi:
                 raise ParseException(
-                    string,
+                    self,
                     start,
+                    string,
                     "Missing minimum (%i) more required elements (%s)" % (mi, e),
                 )
 
@@ -494,11 +762,12 @@ class Each(ParseExpression):
         # add any unmatched Optionals, in case they have default values defined
         matchOrder += [e for e in self.exprs if id(e) not in found]
 
+        # TODO: CAN WE AVOID THIS RE-PARSE?
         results = []
         end = start
         for e in matchOrder:
             result = e._parse(string, end, doActions)
-            end = result.end
+            end = self.parser_config.whitespace.skip(string, result.end)
             results.append(result)
 
         return ParseResults(self, results[0].start, results[-1].end, results)
@@ -510,16 +779,7 @@ class Each(ParseExpression):
         return "{" + " & ".join(text(e) for e in self.exprs) + "}"
 
 
-# export
-from mo_parsing import core, engine
-
-core.And = And
-core.Or = Or
-core.Each = Each
-core.MatchFirst = MatchFirst
-
-from mo_parsing import helpers
-
-helpers.And = And
-helpers.Or = Or
-helpers.MatchFirst = MatchFirst
+export("mo_parsing.core", And)
+export("mo_parsing.core", Or)
+export("mo_parsing.core", MatchAll)
+export("mo_parsing.core", MatchFirst)
