@@ -1,17 +1,26 @@
 # encoding: utf-8
 import inspect
-from collections import MutableMapping
 
 from mo_dots import is_many, is_null
-from mo_future import is_text, text, PY3, NEXT, zip_longest
-from mo_parsing.utils import Log
+from mo_future import is_text, text, NEXT, zip_longest, MutableMapping
+from mo_imports import expect, export
 
-from mo_parsing import engine
+from mo_parsing.utils import Log, listwrap
 from mo_parsing.utils import is_forward, forward_type
 
 USE_ATTRIBUTE_ACCESS = False
 
-Suppress, ParserElement, NO_PARSER, NO_RESULTS, Group, Dict, Token, Empty = [None] * 8
+
+Suppress, ParserElement, NO_PARSER, NO_RESULTS, Group, Dict, Token, Empty = expect(
+    "Suppress",
+    "ParserElement",
+    "NO_PARSER",
+    "NO_RESULTS",
+    "Group",
+    "Dict",
+    "Token",
+    "Empty",
+)
 
 
 class ParseResults(object):
@@ -22,10 +31,13 @@ class ParseResults(object):
         return self.type.token_name
 
     def __init__(self, result_type, start, end, tokens):
+        if end == -1:
+            Log.error("not allowed")
         self.type = result_type
         self.start = start
         self.end = end
         self.tokens = tokens
+        self.timing = None
 
     def _get_item_by_name(self, name):
         # return open list of (modal, value) pairs
@@ -35,22 +47,17 @@ class ParseResults(object):
                 if tok.name == name:
                     if isinstance(tok.type, Group):
                         yield tok
-                    elif is_forward(tok.type) and isinstance(tok.tokens[0].type, Group):
-                        yield tok
                     else:
-                        if any(isinstance(t, ParseResults) and t.name for t in tok.tokens):
-                            for t in tok.tokens:
-                                yield t
-                        else:
-                            for t in tok:
-                                yield t
+                        for t in tok.tokens:
+                            for tt in _flatten(t):
+                                yield tt
+                    continue
+                elif tok.name:
                     continue
                 elif isinstance(tok.type, Group):
                     continue
                 elif is_forward(tok.type) and isinstance(tok.tokens[0].type, Group):
                     continue
-                elif tok.name:
-                    continue
                 for f in tok._get_item_by_name(name):
                     yield f
 
@@ -58,7 +65,15 @@ class ParseResults(object):
         if is_forward(self.type):
             return self.tokens[0][item]
 
-        if isinstance(item, int):
+        if is_text(item):
+            values = list(self._get_item_by_name(item))
+            if len(values) == 0:
+                return NO_RESULTS
+            if len(values) == 1:
+                return values[0]
+            # ENCAPSULATE IN A ParseResults FOR FURTHER NAVIGATION
+            return ParseResults(NO_PARSER, -1, 0, values)
+        elif isinstance(item, int):
             if item < 0:
                 item = len(self) + item
             for ii, v in enumerate(self):
@@ -67,13 +82,7 @@ class ParseResults(object):
         elif isinstance(item, slice):
             return list(iter(self))[item]
         else:
-            values = list(self._get_item_by_name(item))
-            if len(values) == 0:
-                return NO_RESULTS
-            if len(values) == 1:
-                return values[0]
-            # ENCAPSULATE IN A ParseResults FOR FURTHER NAVIGATION
-            return ParseResults(NO_PARSER, -1, -1, values)
+            Log.error("not expected")
 
     def __setitem__(self, k, v):
         if isinstance(k, (slice, int)):
@@ -82,7 +91,7 @@ class ParseResults(object):
         if v is None:
             v = NO_RESULTS
 
-        if is_forward(self):
+        if is_forward(self.type):
             self.tokens[0][k] = v
             return
 
@@ -101,7 +110,13 @@ class ParseResults(object):
                 tok.__setitem__(k, NO_RESULTS)  # ERASE ALL CHILDREN
 
         if v is not NO_RESULTS:
-            self.tokens.append(Annotation(k, -1, -1, [v]))
+            tokens = self.tokens
+            if is_forward(self.type):
+                tokens = tokens[0].tokens
+            if isinstance(v, ParseResults):
+                tokens.append(Annotation(k, v.start, v.end, v.tokens))
+            else:
+                tokens.append(Annotation(k, -1, 0, listwrap(v)))
 
     if USE_ATTRIBUTE_ACCESS:
 
@@ -139,12 +154,17 @@ class ParseResults(object):
             return self[0] == other
         elif not self:
             return False
+        elif isinstance(other, dict):
+            for k, v in other.items():
+                if self[k] != v:
+                    return False
+            return True
         else:
             Log.error("do not know how to handle")
 
     def __bool__(self):
         try:
-            NEXT(self.iteritems())()
+            NEXT(self.items())()
             return True
         except Exception:
             pass
@@ -215,17 +235,46 @@ class ParseResults(object):
     def __reversed__(self):
         return reversed(self.tokens)
 
-    def iterkeys(self):
-        for k, _ in self.iteritems():
+    # def __getattr__(self, item):
+    #     """
+    #     IF THERE IS ONLY ONE VALUE, THEN DEFER TO IT
+    #     """
+    #     iter = self.__iter__()
+    #     try:
+    #         v1 = iter.__next__()
+    #         try:
+    #             iter.__next__()
+    #             raise Log.error("No attribute {{item}} for mutiple tokens", item=item)
+    #         except Exception:
+    #             return getattr(v1, item)
+    #     except Exception as cause:
+    #         raise AttributeError(f"No attribute {item}")
+
+    def value(self):
+        """
+        RETURN WHATEVER PRIMITIVE VALUES ARE LEFT
+        """
+        value = [v.value() if isinstance(v, ParseResults) else v for v in self]
+        if not value:
+            return None
+        elif len(value) == 1:
+            return value[0]
+        else:
+            return value
+
+    def keys(self):
+        for k, _ in self.items():
             yield k
 
-    def itervalues(self):
-        for _, v in self.iteritems():
+    def values(self):
+        for _, v in self.items():
             yield v
 
-    def iteritems(self):
+    def items(self):
         if is_forward(self.type):
-            return self.tokens[0].iteritems()
+            for k, v in self.tokens[0].items():
+                yield k, v
+            return
 
         output = {}
         for tok in self.tokens:
@@ -237,26 +286,11 @@ class ParseResults(object):
                     continue
                 if is_forward(tok.type) and isinstance(tok.tokens[0].type, Group):
                     continue
-                for k, v in tok.iteritems():
+                for k, v in tok.items():
                     add(output, k, v)
         for k, v in output.items():
             yield k, v
 
-    if PY3:
-        keys = iterkeys
-        values = itervalues
-        items = iteritems
-    else:
-
-        def keys(self):
-            return list(self.iterkeys())
-
-        def values(self):
-            return list(self.itervalues())
-
-        def items(self):
-            return list(self.iteritems())
-
     def haskeys(self):
         """Since keys() returns an iterator, this method is helpful in bypassing
         code that looks for the existence of any defined results names."""
@@ -290,9 +324,14 @@ class ParseResults(object):
             return defaultValue
 
     def __contains__(self, item):
+        if item is Ellipsis:
+            return False
         return bool(self[item])
 
     def __add__(self, other):
+        # if not isinstance(other, ParseResults):
+        #     return self.value() + other
+        #
         return ParseResults(
             Group(self.type + other.type),
             self.start,
@@ -303,7 +342,7 @@ class ParseResults(object):
     def __radd__(self, other):
         if not other:  # happens when using sum() on parsers
             return self
-        other = engine.CURRENT.normalize(other)
+        other = whitespaces.CURRENT.normalize(other)
         return other + self
 
     def __repr__(self):
@@ -417,7 +456,7 @@ class ParseResults(object):
         elif len(self.tokens) == 1:
             return self.tokens[0].name
         else:
-            return None
+            return ""
 
     def __getnewargs__(self):
         old_parser = self.type
@@ -430,14 +469,20 @@ class ParseResults(object):
         return dir(type(self))
 
 
-def simpler(v):
-    # convert an open list to object it represents
-    if isinstance(v, list):
-        if len(v) == 0:
-            return None
-        elif len(v) == 1:
-            return v[0]
-    return v
+def _flatten(token):
+    """
+    FLATTEN SOME, LEAVING ANY IMPORTANT FEATURES (names or groups)
+    """
+    if not isinstance(token, ParseResults):
+        yield token
+    elif isinstance(token.type, Group):
+        yield token
+    elif token.name:
+        yield token
+    else:
+        for t in token.tokens:
+            for tt in _flatten(t):
+                yield tt
 
 
 def add(obj, key, value):
@@ -473,3 +518,5 @@ class Annotation(ParseResults):
 
 
 MutableMapping.register(ParseResults)
+
+export("mo_parsing.utils", ParseResults)
