@@ -1,29 +1,21 @@
 # encoding: utf-8
-import re
-import sre_constants
-import warnings
 
-from mo_future import text
+from mo_future import is_text
+from mo_imports import export
 
+from mo_parsing import whitespaces
 from mo_parsing.core import ParserElement
-from mo_parsing.engine import Engine, PLAIN_ENGINE
+from mo_parsing.whitespaces import Whitespace
 from mo_parsing.exceptions import ParseException
 from mo_parsing.results import ParseResults
-from mo_parsing.utils import (
-    Log,
-    regex_range,
-    append_config,
-    regex_type,
-    regex_caseless,
-)
-from mo_parsing.utils import (
-    MAX_INT,
-    col,
-    printables,
-)
+from mo_parsing.utils import *
 
 
 class Token(ParserElement):
+    """
+    Represent some contiguous set for characters, no whitespace
+    """
+
     __slots__ = []
     Config = append_config(ParserElement, "match", "regex")
 
@@ -38,6 +30,7 @@ class Empty(Token):
     Often used to consume-and-suppress trailing whitespace
     """
 
+    zero_length = True
     __slots__ = []
 
     def __init__(self, name=None):
@@ -51,11 +44,15 @@ class Empty(Token):
         return 0
 
     def parseImpl(self, string, start, doActions=True):
-        end = self.engine.skip(string, start)
+        end = start
+        # end = self.whitespace.skip(string, start)
         return ParseResults(self, start, end, [])
 
+    def streamline(self):
+        return self
+
     def __regex__(self):
-        return self.engine.__regex__()
+        return self.whitespace.__regex__()
 
     def __str__(self):
         return self.parser_name or "Empty"
@@ -64,6 +61,7 @@ class Empty(Token):
 class NoMatch(Token):
     """A token that will never match."""
 
+    zero_length = True
     __slots__ = []
 
     def __init__(self):
@@ -80,6 +78,24 @@ class NoMatch(Token):
         return "+", "a^"
 
 
+class LookBehind(Token):
+    zero_length = True
+    __slots__ = []
+
+    def __init__(self, name=None):
+        Token.__init__(self)
+        self.parser_name = name
+
+    def is_annotated(self):
+        return self.parseAction or self.token_name
+
+    def min_length(self):
+        return 0
+
+    def streamline(self):
+        return self
+
+
 class AnyChar(Token):
     __slots__ = []
 
@@ -108,7 +124,10 @@ class Literal(Token):
     __slots__ = []
 
     def __init__(self, match):
+        if not is_text(match):
+            Log.error("Expecting string for literal")
         Token.__init__(self)
+
         self.set_config(match=match)
 
         if len(match) == 0:
@@ -123,6 +142,9 @@ class Literal(Token):
             return ParseResults(self, start, end, [match])
         raise ParseException(self, start, string)
 
+    def expecting(self):
+        return {self.parser_config.match.lower(): [self]}
+
     def _min_length(self):
         return len(self.parser_config.match)
 
@@ -153,33 +175,13 @@ class SingleCharLiteral(Literal):
 
 
 class Keyword(Token):
-    """Token to exactly match a specified string as a keyword, that is,
-    it must be immediately followed by a non-keyword character.  Compare
-    with `Literal`:
-
-     - ``Literal("if")`` will match the leading ``'if'`` in
-       ``'ifAndOnlyIf'``.
-     - ``Keyword("if")`` will not; it will only match the leading
-       ``'if'`` in ``'if x=1'``, or ``'if(y==2)'``
-
-    Accepts two optional constructor arguments in addition to the
-    keyword string:
-
-     - ``ident_chars`` is a string of characters that would be valid
-       identifier characters, defaulting to all alphanumerics + "_" and
-       "$"
-     - ``caseless`` allows case-insensitive matching, default is ``False``.
-
-    For case-insensitive matching, use `CaselessKeyword`.
-    """
-
     __slots__ = []
     Config = append_config(Token, "ident_chars")
 
     def __init__(self, match, ident_chars=None, caseless=None):
         Token.__init__(self)
         if ident_chars is None:
-            ident_chars = self.engine.keyword_chars
+            ident_chars = whitespaces.CURRENT.keyword_chars
         else:
             ident_chars = "".join(sorted(set(ident_chars)))
 
@@ -190,7 +192,9 @@ class Keyword(Token):
 
         non_word = "($|(?!" + regex_range(ident_chars) + "))"
         self.set_config(
-            ident_chars=ident_chars, match=match, regex=re.compile(pattern + non_word)
+            ident_chars=ident_chars,
+            match=match,
+            regex=regex_compile(pattern + non_word),
         )
 
         self.parser_name = match
@@ -203,6 +207,9 @@ class Keyword(Token):
             return ParseResults(self, start, found.end(), [self.parser_config.match])
         raise ParseException(self, start, string)
 
+    def expecting(self):
+        return {self.parser_config.match.lower(): [self]}
+
     def _min_length(self):
         return len(self.parser_config.match)
 
@@ -228,8 +235,7 @@ class CaselessLiteral(Literal):
     def __init__(self, match):
         Literal.__init__(self, match.upper())
         self.set_config(
-            match=match,
-            regex=re.compile(regex_caseless(match), re.MULTILINE | re.DOTALL),
+            match=match, regex=regex_compile(regex_caseless(match)),
         )
         self.parser_name = repr(self.parser_config.regex.pattern)
 
@@ -299,45 +305,8 @@ class CloseMatch(Token):
 
 
 class Word(Token):
-    """Token for matching words composed of allowed character sets.
-    Defined with string containing all allowed initial characters, an
-    optional string containing allowed body characters (if omitted,
-    defaults to the initial character set), and an optional minimum,
-    maximum, and/or exact length.  The default value for ``min`` is
-    1 (a minimum value < 1 is not valid); the default values for
-    ``max`` and ``exact`` are 0, meaning no maximum or exact
-    length restriction. An optional ``excludeChars`` parameter can
-    list characters that might be found in the input ``body_chars``
-    string; useful to define a word of all printables except for one or
-    two characters, for instance.
-
-    `srange` is useful for defining custom character set strings
-    for defining ``Word`` expressions, using range notation from
-    regular expression character sets.
-
-    A common mistake is to use `Word` to match a specific literal
-    string, as in ``Word("Address")``. Remember that `Word`
-    uses the string argument to define *sets* of matchable characters.
-    This expression would match "Add", "AAA", "dAred", or any other word
-    made up of the characters 'A', 'd', 'r', 'e', and 's'. To match an
-    exact literal string, use `Literal` or `Keyword`.
-
-    mo_parsing includes helper strings for building Words:
-
-     - `alphas`
-     - `nums`
-     - `alphanums`
-     - `hexnums`
-     - `alphas8bit` (alphabetic characters in ASCII range 128-255
-       - accented, tilded, umlauted, etc.)
-     - `punc8bit` (non-alphabetic characters in ASCII range
-       128-255 - currency, symbols, superscripts, diacriticals, etc.)
-     - `printables` (any non-whitespace character)
-
-    """
-
-    __slots__ = []
-    Config = append_config(Token, "min")
+    __slots__ = ["regex"]
+    Config = append_config(Token, "min", "init_chars")
 
     def __init__(
         self,
@@ -346,8 +315,8 @@ class Word(Token):
         min=1,
         max=None,
         exact=0,
-        asKeyword=False,
-        excludeChars=None,
+        as_keyword=False,  # IF WE EXPECT NON-WORD CHARACTERS BEFORE AND AFTER
+        exclude="",
     ):
         Token.__init__(self)
 
@@ -363,62 +332,76 @@ class Word(Token):
             )
 
         if body_chars == init_chars:
-            prec, regexp = Char(
-                init_chars, excludeChars=excludeChars
-            )[min:max].__regex__()
+            prec, regexp = Char(init_chars, exclude=exclude)[min:max].__regex__()
         elif max is None or max == MAX_INT:
             prec, regexp = (
-                Char(init_chars, excludeChars=excludeChars)
-                + Char(body_chars, excludeChars=excludeChars)[min - 1 :]
+                Char(init_chars, exclude=exclude)
+                + Char(body_chars, exclude=exclude)[min - 1 :]
             ).__regex__()
         else:
             prec, regexp = (
-                Char(init_chars, excludeChars=excludeChars)
-                + Char(body_chars, excludeChars=excludeChars)[min - 1 : max - 1]
+                Char(init_chars, exclude=exclude)
+                + Char(body_chars, exclude=exclude)[min - 1 : max - 1]
             ).__regex__()
 
-        if asKeyword:
+        if as_keyword:
             regexp = r"\b" + regexp + r"\b"
 
-        self.set_config(regex=re.compile(regexp, re.MULTILINE | re.DOTALL), min=min)
+        self.regex = regex_compile(regexp)
+        self.set_config(min=min, init_chars=init_chars)
+
+    def copy(self):
+        output = Token.copy(self)
+        output.regex = self.regex
+        return output
 
     def parseImpl(self, string, start, doActions=True):
-        found = self.parser_config.regex.match(string, start)
+        found = self.regex.match(string, start)
         if found:
             return ParseResults(self, start, found.end(), [found.group()])
-
-        raise ParseException(self, start, string)
+        else:
+            raise ParseException(self, start, string)
 
     def min_length(self):
         return self.parser_config.min
 
+    def expecting(self):
+        if len(self.parser_config.init_chars) < 11:
+            return {c: [self] for c in self.parser_config.init_chars}
+        else:
+            return {}
+
     def __regex__(self):
-        return "+", self.parser_config.regex.pattern
+        return "+", self.regex.pattern
 
     def __str__(self):
         if self.parser_name:
             return self.parser_name
-        return f"W:({self.parser_config.regex.pattern})"
+        return f"W:({self.regex.pattern})"
 
 
 class Char(Token):
     __slots__ = []
-    Config = append_config(Token, "charset")
+    Config = append_config(Token, "include", "exclude")
 
-    def __init__(self, charset, asKeyword=False, excludeChars=None):
+    def __init__(self, include="", asKeyword=False, exclude=""):
         """
-        Represent one character in a given charset
+        Represent one character in a given include
         """
         Token.__init__(self)
-        if excludeChars:
-            charset = set(charset) - set(excludeChars)
-        regex = regex_range(charset)
+        include = set(include) if include else set()
+        exclude = set(exclude) if exclude else set()
+        include = "".join(sorted(include - exclude))
+        exclude = "".join(sorted(exclude))
+
+        if not include:
+            regex = regex_range(exclude, exclude=True)
+        else:
+            regex = regex_range(include)
+
         if asKeyword:
             regex = r"\b%s\b" % self
-        self.set_config(
-            regex=re.compile(regex, re.MULTILINE | re.DOTALL),
-            charset="".join(sorted(set(charset))),
-        )
+        self.set_config(regex=regex_compile(regex), include=include, exclude=exclude)
 
     def parseImpl(self, string, start, doActions=True):
         found = self.parser_config.regex.match(string, start)
@@ -427,6 +410,9 @@ class Char(Token):
 
         raise ParseException(self, start, string)
 
+    def expecting(self):
+        return {c: [self] for c in self.parser_config.include}
+
     def min_length(self):
         return 1
 
@@ -437,294 +423,6 @@ class Char(Token):
         return self.parser_config.regex.pattern
 
 
-class Regex(Token):
-    r"""Token for matching strings that match a given regular
-    expression. Defined with string specifying the regular expression in
-    a form recognized by the stdlib Python  `re module <https://docs.python.org/3/library/re.html>`_.
-    If the given regex contains named groups (defined using ``(?P<name>...)``),
-    these will be preserved as named parse results.
-    """
-    __slots__ = []
-    Config = append_config(Token, "flags")
-
-    def __new__(cls, pattern, flags=0, asGroupList=False, asMatch=False):
-        if asGroupList:
-            return object.__new__(_RegExAsGroup)
-        elif asMatch:
-            return object.__new__(_RegExAsMatch)
-        else:
-            return object.__new__(cls)
-
-    def __init__(self, pattern, flags=0, asGroupList=False, asMatch=False):
-        """The parameters ``pattern`` and ``flags`` are passed
-        to the ``re.compile()`` function as-is. See the Python
-        `re module <https://docs.python.org/3/library/re.html>`_ module for an
-        explanation of the acceptable patterns and flags.
-        """
-        Token.__init__(self)
-
-        if isinstance(pattern, text):
-            if not pattern:
-                warnings.warn(
-                    "null string passed to Regex; use Empty() instead",
-                    SyntaxWarning,
-                    stacklevel=2,
-                )
-
-            try:
-                self.set_config(flags=flags, regex=re.compile(pattern, flags))
-            except sre_constants.error as cause:
-                Log.error(
-                    "invalid pattern {{pattern}} passed to Regex",
-                    pattern=pattern,
-                    cause=cause,
-                )
-        elif isinstance(pattern, regex_type):
-            self.set_config(flags=flags, regex=pattern)
-        else:
-            Log.error(
-                "Regex may only be constructed with a string or a compiled RE object"
-            )
-
-        self.parser_name = text(self)
-
-    def parseImpl(self, string, start, doActions=True):
-        found = self.parser_config.regex.match(string, start)
-        if found:
-            ret = ParseResults(self, start, found.end(), [found.group()])
-            d = found.groupdict()
-            if d:
-                for k, v in d.items():
-                    ret[k] = v
-            return ret
-
-        raise ParseException(self, start, string)
-
-    def min_length(self):
-        return 0
-
-    def __regex__(self):
-        return "|", self.parser_config.regex.pattern
-
-    def __str__(self):
-        return self.parser_config.regex.pattern
-
-    def sub(self, repl):
-        r"""
-        Return Regex with an attached parse action to transform the parsed
-        result as if called using `re.sub(expr, repl, string) <https://docs.python.org/3/library/re.html#re.sub>`_.
-
-        Example::
-
-            make_html = Regex(r"(\w+):(.*?):").sub(r"<\1>\2</\1>")
-            print(make_html.transformString("h1:main title:"))
-            # prints "<h1>main title</h1>"
-        """
-
-        def pa(tokens):
-            return self.parser_config.regex.sub(repl, tokens[0])
-
-        return self.addParseAction(pa)
-
-
-class _RegExAsGroup(Regex):
-    __slots__ = []
-
-    def parseImpl(self, string, start, doActions=True):
-        result = self.parser_config.regex.match(string, start)
-        if not result:
-            raise ParseException(self, start, string)
-
-        return ParseResults(self, start, result.end(), [result.groups()])
-
-    def sub(self, repl):
-        raise SyntaxError("cannot use sub() with Regex(asGroupList=True)")
-
-
-class _RegExAsMatch(Regex):
-    __slots__ = []
-
-    def parseImpl(self, string, start, doActions=True):
-        result = self.parser_config.regex.match(string, start)
-        if not result:
-            raise ParseException(self, start, string)
-
-        return ParseResults(self, start, result.end(), [result])
-
-    def sub(self, repl):
-        if callable(repl):
-            raise SyntaxError(
-                "cannot use sub() with a callable with Regex(asMatch=True)"
-            )
-
-        def pa(tokens):
-            return tokens[0].expand(repl)
-
-        return self.addParseAction(pa)
-
-
-class QuotedString(Token):
-    r"""
-    Token for matching strings that are delimited by quoting characters.
-
-    Defined with the following parameters:
-
-        - quoteChar - string of one or more characters defining the
-          quote delimiting string
-        - escChar - character to escape quotes, typically backslash
-          (default= ``None``)
-        - escQuote - special quote sequence to escape an embedded quote
-          string (such as SQL's ``""`` to escape an embedded ``"``)
-          (default= ``None``)
-        - multiline - boolean indicating whether quotes can span
-          multiple lines (default= ``False``)
-        - unquoteResults - boolean indicating whether the matched text
-          should be unquoted (default= ``True``)
-        - endQuoteChar - string of one or more characters defining the
-          end of the quote delimited string (default= ``None``  => same as
-          quoteChar)
-        - convertWhitespaceEscapes - convert escaped whitespace
-          (``'\t'``, ``'\n'``, etc.) to actual whitespace
-          (default= ``True``)
-
-    """
-    __slots__ = []
-    Config = append_config(
-        Token,
-        "quote_char",
-        "end_quote_char",
-        "esc_char",
-        "esc_quote",
-        "unquoteResults",
-        "convertWhitespaceEscapes",
-        "flags",
-        "escCharReplacePattern",
-    )
-
-    def __init__(
-        self,
-        quoteChar,
-        escChar=None,
-        escQuote=None,
-        multiline=False,
-        unquoteResults=True,
-        endQuoteChar=None,
-        convertWhitespaceEscapes=True,
-    ):
-        super(QuotedString, self).__init__()
-
-        # remove white space from quote chars - wont work anyway
-        quoteChar = quoteChar.strip()
-        if not quoteChar:
-            warnings.warn(
-                "quoteChar cannot be the empty string", SyntaxWarning, stacklevel=2
-            )
-            raise SyntaxError()
-
-        if endQuoteChar is None:
-            endQuoteChar = quoteChar
-        else:
-            endQuoteChar = endQuoteChar.strip()
-            if not endQuoteChar:
-                warnings.warn(
-                    "endQuoteChar cannot be the empty string",
-                    SyntaxWarning,
-                    stacklevel=2,
-                )
-                raise SyntaxError()
-
-        self.set_config(
-            quote_char=quoteChar,
-            end_quote_char=endQuoteChar,
-            esc_char=escChar,
-            esc_quote=escQuote,
-            unquoteResults=unquoteResults,
-            convertWhitespaceEscapes=convertWhitespaceEscapes,
-        )
-        # TODO: FIX THIS MESS. WE SHOULD BE ABLE TO CONSTRUCT REGEX FROM ParserElements
-        included = Empty()
-        excluded = Literal(self.parser_config.end_quote_char)
-
-        if multiline:
-            flags = re.MULTILINE | re.DOTALL
-        else:
-            excluded |= Char("\r\n")
-            flags = 0
-        if escQuote:
-            included |= Literal(escQuote)
-        if escChar:
-            excluded |= Literal(self.parser_config.esc_char)
-            included = included | escChar + Char(printables)
-            self.set_config(
-                escCharReplacePattern=re.escape(self.parser_config.esc_char) + "(.)"
-            )
-
-        prec, pattern = (
-            Literal(quoteChar)
-            + ((~excluded + AnyChar()) | included)[0:]
-            + Literal(self.parser_config.end_quote_char)
-        ).__regex__()
-
-        self.set_config(flags=flags, regex=re.compile(pattern, flags))
-
-        self.parser_name = text(self)
-
-    def parseImpl(self, string, start, doActions=True):
-        found = self.parser_config.regex.match(string, start)
-        if not found:
-            raise ParseException(self, start, string)
-
-        end = found.end()
-        ret = found.group()
-
-        if self.parser_config.unquoteResults:
-
-            # strip off quotes
-            ret = ret[
-                len(self.parser_config.quote_char) : -len(self.parser_config.end_quote_char)
-            ]
-
-            if isinstance(ret, text):
-                # replace escaped whitespace
-                if "\\" in ret and self.parser_config.convertWhitespaceEscapes:
-                    ws_map = {
-                        r"\t": "\t",
-                        r"\n": "\n",
-                        r"\f": "\f",
-                        r"\r": "\r",
-                    }
-                    for wslit, wschar in ws_map.items():
-                        ret = ret.replace(wslit, wschar)
-
-                # replace escaped characters
-                if self.parser_config.esc_char:
-                    ret = re.sub(
-                        self.parser_config.escCharReplacePattern, r"\g<1>", ret
-                    )
-
-                # replace escaped quotes
-                if self.parser_config.esc_quote:
-                    ret = ret.replace(
-                        self.parser_config.esc_quote, self.parser_config.end_quote_char
-                    )
-
-        return ParseResults(self, start, end, [ret])
-
-    def min_length(self):
-        return 2
-
-    def __str__(self):
-        try:
-            return super(QuotedString, self).__str__()
-        except Exception:
-            pass
-
-        return "quoted string, starting with %s ending with %s" % (
-            self.parser_config.quote_char,
-            self.parser_config.end_quote_char,
-        )
-
-
 class CharsNotIn(Token):
     """Token for matching words composed of characters *not* in a given
     set (will include whitespace in matched characters if not listed in
@@ -772,7 +470,7 @@ class CharsNotIn(Token):
             suffix = "{" + str(min) + ":" + str(max) + "}"
 
         self.set_config(
-            regex=re.compile(regex + suffix),
+            regex=regex_compile(regex + suffix),
             min_len=min,
             max_len=max,
             not_chars=not_chars,
@@ -797,7 +495,8 @@ class CharsNotIn(Token):
 
 
 class White(Token):
-    """Special matching class for matching whitespace.  Normally,
+    """
+    Special matching class for matching whitespace.  Normally,
     whitespace is ignored by mo_parsing grammars.  This class is included
     when some whitespace structures are significant.  Define with
     a string containing the whitespace characters to be matched; default
@@ -836,11 +535,7 @@ class White(Token):
     Config = append_config(Token, "min_len", "max_len", "white_chars")
 
     def __init__(self, ws=" \t\r\n", min=1, max=0, exact=0):
-        with Engine(white="".join(
-            c for c in self.engine.white_chars if c not in ws
-        )) as e:
-            super(White, self).__init__()
-            self.set_config(lock_engine=e)
+        Token.__init__(self)
         white_chars = "".join(sorted(set(ws)))
         self.parser_name = "|".join(White.whiteStrs[c] for c in white_chars)
 
@@ -866,80 +561,43 @@ class White(Token):
         return ParseResults(self, start, end, string[start:end])
 
 
-class _PositionToken(Token):
-    __slots__ = []
-
-    def __init__(self):
-        super(_PositionToken, self).__init__()
-        self.parser_name = self.__class__.__name__
-
-    def min_length(self):
-        return 0
-
-
-class GoToColumn(_PositionToken):
-    """Token to advance to a specific column of input text; useful for
-    tabular report scraping.
-    """
-
-    __slots__ = []
-
-    def __init__(self, colno):
-        super(GoToColumn, self).__init__()
-        self.col = colno
-
-    def preParse(self, string, loc):
-        if col(loc, string) != self.col:
-            instrlen = len(string)
-            loc = self._skipIgnorables(string, loc)
-            while (
-                loc < instrlen
-                and string[loc].isspace()
-                and col(loc, string) != self.col
-            ):
-                loc += 1
-        return loc
-
-    def parseImpl(self, string, start, doActions=True):
-        thiscol = col(start, string)
-        if thiscol > self.col:
-            raise ParseException(self, start, string, "Text not in expected column")
-        newloc = start + self.col - thiscol
-        ret = string[start:newloc]
-        return newloc, ret
-
-
-class LineStart(_PositionToken):
-    r"""Matches if current position is at the beginning of a line within
+class LineStart(Token):
+    r"""
+    Matches if current position is at the beginning of a line within
     the parse string
     """
+    zero_length = True
     __slots__ = []
 
     def __init__(self):
-        super(LineStart, self).__init__()
+        Token.__init__(self)
+        self.parser_name = self.__class__.__name__
 
     def parseImpl(self, string, start, doActions=True):
         if col(start, string) == 1:
             return ParseResults(self, start, start, [])
         raise ParseException(self, start, string)
 
-    def __regeex__(self):
-        return "^"
+    def min_length(self):
+        return 0
+
+    def __regex__(self):
+        return "*", "^"
 
 
-class LineEnd(_PositionToken):
-    """Matches if current position is at the end of a line within the
-    parse string
+class LineEnd(LookBehind):
+    """
+    Matches if current position is at the end of a line
     """
 
+    zero_length = True
     __slots__ = []
 
     def __init__(self):
-        with Engine(" \t") as e:
-            super(LineEnd, self).__init__()
-            self.set_config(
-                lock_engine=e, regex=re.compile("\\r?(\\n|$)", re.MULTILINE | re.DOTALL)
-            )
+        with Whitespace(" \t") as e:
+            Token.__init__(self)
+            self.parser_name = self.__class__.__name__
+            self.set_config(lock_engine=e, regex=regex_compile("\\r?(\\n|$)"))
 
     def parseImpl(self, string, start, doActions=True):
         found = self.parser_config.regex.match(string, start)
@@ -954,46 +612,59 @@ class LineEnd(_PositionToken):
         return "|", self.parser_config.regex.pattern
 
 
-class StringStart(_PositionToken):
+class StringStart(Token):
     """Matches if current position is at the beginning of the parse
     string
     """
 
+    zero_length = True
     __slots__ = []
 
     def __init__(self):
-        super(StringStart, self).__init__()
+        Token.__init__(self)
+        self.parser_name = self.__class__.__name__
 
     def parseImpl(self, string, loc, doActions=True):
         if loc != 0:
             # see if entire string up to here is just whitespace and ignoreables
-            if loc != self.engine.skip(string, 0):
-                raise ParseException(self, loc, string)
+            # if loc != self.whitespace.skip(string, 0):
+            raise ParseException(self, loc, string)
         return []
 
+    def min_length(self):
+        return 0
 
-class StringEnd(_PositionToken):
+
+class StringEnd(Token):
     """
     Matches if current position is at the end of the parse string
     """
 
+    zero_length = True
     __slots__ = []
 
     def __init__(self):
-        with Engine() as e:
-            super(StringEnd, self).__init__()
-            self.set_config(lock_engine=e)
+        Token.__init__(self)
+        self.parser_name = self.__class__.__name__
+        self.set_config(regex=regex_compile("(?:\\r?\\n)*$"))
 
     def parseImpl(self, string, start, doActions=True):
         end = len(string)
         if start >= end:
             return ParseResults(self, end, end, [])
 
+        found = self.parser_config.regex.match(string, start)
+        if found:
+            return ParseResults(self, start, found.end(), [])
         raise ParseException(self, start, string)
 
+    def min_length(self):
+        return 0
 
-class WordStart(_PositionToken):
-    """Matches if the current position is at the beginning of a Word,
+
+class WordStart(Token):
+    """
+    Matches if the current position is at the beginning of a Word,
     and is not preceded by any character in a given set of
     ``wordChars`` (default= ``printables``). To emulate the
     ``\b`` behavior of regular expressions, use
@@ -1002,15 +673,16 @@ class WordStart(_PositionToken):
     a line.
     """
 
+    zero_length = True
     __slots__ = []
-    Config = append_config(_PositionToken, "word_chars")
+    Config = append_config(Token, "word_chars")
 
     def __init__(self, wordChars=printables):
-        super(WordStart, self).__init__()
+        Token.__init__(self)
+        self.parser_name = self.__class__.__name__
         self.set_config(
-            regex=re.compile(
-                f"(?:(?<={(CharsNotIn(wordChars, exact=1)).__regex__()[1]})|^)(?={Char(wordChars).__regex__()[1]})",
-                re.MULTILINE | re.DOTALL,
+            regex=regex_compile(
+                f"(?:(?<={(CharsNotIn(wordChars, exact=1)).__regex__()[1]})|^)(?={Char(wordChars).__regex__()[1]})"
             ),
             word_chars="".join(sorted(set(wordChars))),
         )
@@ -1029,8 +701,9 @@ class WordStart(_PositionToken):
         return "+", self.parser_config.regex.pattern
 
 
-class WordEnd(_PositionToken):
-    """Matches if the current position is at the end of a Word, and is
+class WordEnd(Token):
+    """
+    Matches if the current position is at the end of a Word, and is
     not followed by any character in a given set of ``wordChars``
     (default= ``printables``). To emulate the ``\b`` behavior of
     regular expressions, use ``WordEnd(alphanums)``. ``WordEnd``
@@ -1038,62 +711,57 @@ class WordEnd(_PositionToken):
     of a line.
     """
 
+    zero_length = True
     __slots__ = []
-    Config = append_config(_PositionToken, "word_chars")
+    Config = append_config(Token, "word_chars")
 
-    def __init__(self, wordChars=printables):
-        super(WordEnd, self).__init__()
-        self.engine = PLAIN_ENGINE
+    def __init__(self, word_chars=None):
+        Token.__init__(self)
+        word_chars = coalesce(word_chars, whitespaces.CURRENT.keyword_chars)
+        self.parser_name = self.__class__.__name__
         self.set_config(
-            word_chars="".join(sorted(set(wordChars))),
-            regex=re.compile(
-                f"(?<={Char(wordChars).__regex__()[1]})({(~Char(wordChars)).__regex__()[1]}|$)",
-                re.MULTILINE | re.DOTALL,
+            word_chars="".join(sorted(set(word_chars))),
+            regex=regex_compile(
+                f"(?<={Char(word_chars).__regex__()[1]})({(~Char(word_chars)).__regex__()[1]}|$)"
             ),
         )
 
     def copy(self):
-        output = _PositionToken.copy(self)
-        output.engine = PLAIN_ENGINE
+        output = Token.copy(self)
         return output
 
+    def parseImpl(self, string, start, doActions=True):
+        found = self.parser_config.regex.match(string, start)
+        if found:
+            return ParseResults(self, start, start, [])
+        else:
+            raise ParseException(self, start, string)
+
     def min_length(self):
         return 0
 
-    def parseImpl(self, string, start, doActions=True):
-        word_chars = self.parser_config.word_chars
-        instrlen = len(string)
-        if instrlen > 0 and start < instrlen:
-            if string[start] in word_chars or string[start - 1] not in word_chars:
-                raise ParseException(self, start, string)
-        return ParseResults(self, start, start, [])
-
     def __regex__(self):
         return "+", self.parser_config.regex.pattern
 
 
-# export
-from mo_parsing import core, enhancement, engine, results
+export("mo_parsing.results", Token)
+export("mo_parsing.results", Empty)
 
-core.Empty = Empty
-core.StringEnd = StringEnd
-core.Literal = Literal
-core.Token = Token
+export("mo_parsing.core", Empty)
+export("mo_parsing.core", StringEnd)
+export("mo_parsing.core", Literal)
+export("mo_parsing.core", Token)
 
-engine.Token = Token
-engine.Literal = Literal
-engine.CURRENT.literal = Literal
+export("mo_parsing.whitespaces", Literal)
+export("mo_parsing.whitespaces", Token)
 
-enhancement.Token = Token
-enhancement.Literal = Literal
-enhancement.Keyword = Keyword
-enhancement.Word = Word
-enhancement.CharsNotIn = CharsNotIn
-enhancement._PositionToken = _PositionToken
-enhancement.StringEnd = StringEnd
-enhancement.Empty = Empty
-enhancement.NoMatch = NoMatch
-enhancement.Char = Char
-
-results.Token = Token
-results.Empty = Empty
+export("mo_parsing.enhancement", Token)
+export("mo_parsing.enhancement", NoMatch)
+export("mo_parsing.enhancement", Literal)
+export("mo_parsing.enhancement", Keyword)
+export("mo_parsing.enhancement", Word)
+export("mo_parsing.enhancement", CharsNotIn)
+export("mo_parsing.enhancement", StringEnd)
+export("mo_parsing.enhancement", Empty)
+export("mo_parsing.enhancement", Char)
+export("mo_parsing.enhancement", LookBehind)
