@@ -13,18 +13,22 @@ from types import FunctionType
 
 from mo_dots import is_null, Null
 from mo_future import unichr, text, generator_types, get_function_name
-from mo_imports import delay_import
+from mo_imports import expect
 
-ParseException = delay_import("mo_parsing.exceptions.ParseException")
+ParseResults, ParseException, Many = expect("ParseResults", "ParseException", "Many")
 
 
 def append_config(base, *slots):
+    dups = set(slots) & set(base.Config._fields)
+    if dups:
+        Log.error("Duplicate config fields: {{dups}}", dups=dups)
+
     fields = base.Config._fields + slots
     return namedtuple("Config", fields)
 
 
 try:
-    from mo_logs import Log
+    from mo_logs import Log, Except
 except Exception:
 
     class Log(object):
@@ -67,7 +71,16 @@ def extend(cls):
 
 _prec = {"|": 0, "+": 1, "*": 2}
 
-regex_type = type(re.compile("[A-Z]", re.MULTILINE | re.DOTALL))
+
+def regex_compile(pattern):
+    """REGEX COMPILE WITHOUT THE ON-A-SINGLE-LINE ASSUMPTION"""
+    try:
+        return re.compile(pattern, re.DOTALL)
+    except Exception as cause:
+        Log.error("could not compile {{pattern}}", pattern=pattern, cause=cause)
+
+
+regex_type = type(regex_compile("[A-Z]"))
 
 
 def regex_iso(curr_prec, expr, new_prec):
@@ -92,29 +105,25 @@ def regex_caseless(literal):
     )
 
 
-_escapes = {
-    "\\": "\\\\",
-    "^": "\\^",
-    "-": "\\-",
-    "]": "\\]",
-    "\n": "\\n",
-    "\r": "\\r",
-    "\t": "\\t",
-}
+_escapes = {"\n": "\\n", "\r": "\\r", "\t": "\\t"}
+_escapes.update({c: "\\" + c for c in r".^$*?+-{}[]\|()"})
 
 
-def regex_range(s):
+def regex_range(s, exclude=False):
     def esc(s):
         return _escapes.get(s, s)
 
     if not s:
         return ""
     if len(s) == 1:
-        return re.escape(s)
+        if exclude:
+            return f"[^{esc(s)}]"
+        else:
+            return esc(s)
 
     start = None
     prev = None
-    acc = ["["]
+    acc = ["[^"] if exclude else ["["]
     for c in list(sorted(set(s))) + ["\a"]:
         if prev and ord(prev) == ord(c) - 1:
             if not start:
@@ -228,6 +237,16 @@ def is_forward(expr):
     return expr.__class__.__name__ == "Forward"
 
 
+def is_backtracking(expr):
+    """
+    RETURN true IF THIS CAN BE EXPENSIVE BACKTRACKER
+    """
+    return (
+        isinstance(expr, Many)
+        and expr.parser_config.max_match - expr.parser_config.min_match > 3
+    )
+
+
 def forward_type(expr):
     """
     :param expr:
@@ -326,9 +345,6 @@ def line(loc, string):
 
 
 def wrap_parse_action(func):
-    from mo_parsing.exceptions import ParseException
-    from mo_parsing.results import ParseResults
-
     if func in singleArgBuiltins:
         spec = inspect.getfullargspec(func)
     elif func.__class__.__name__ == "staticmethod":
@@ -352,12 +368,14 @@ def wrap_parse_action(func):
         num_args = len(spec.args)
 
     def wrapper(*args):
-        token, index, string = args
         try:
+            token, index, string = args
             result = func(*args[:num_args])
             if result is None:
                 return token
             elif isinstance(result, ParseResults):
+                if (result.start < token.start) or (token.end < result.end):
+                    Log.error("Tokens must be ordered")
                 return result
 
             if isinstance(result, (list, tuple)):
